import ast
import functools
import re

from .clike import CLikeTranspiler
from .declaration_extractor import DeclarationExtractor
from .inference import get_inferred_kotlin_type
from py2many.tracer import is_list, defined_before, is_class_or_module, is_self_arg

from py2many.analysis import get_id, is_mutable, is_void_function
from typing import Optional, List, Tuple


class KotlinPrintRewriter(ast.NodeTransformer):
    def __init__(self):
        super().__init__()
        self._temp = 0

    def _get_temp(self):
        self._temp += 1
        return f"__tmp{self._temp}"

    def visit_Call(self, node):
        fname = self.visit(node.func)
        if (
            get_id(fname) == "print"
            and len(node.args) == 1
            and not (
                isinstance(node.args[0], ast.Name)
                or isinstance(node.args[0], ast.Constant)
            )
        ):
            tmp = ast.Name(id=self._get_temp(), lineno=node.lineno)
            ret = ast.If(
                test=ast.Constant(value=True),
                body=[
                    ast.Assign(targets=[tmp], value=node.args[0], lineno=node.lineno),
                    node,
                ],
                orelse=[],
                lineno=node.lineno,
            )
            node.args[0] = tmp
            return ret

        return node


class KotlinTranspiler(CLikeTranspiler):
    NAME = "kotlin"

    CONTAINER_TYPE_MAP = {
        "List": "Array",
        "Dict": "Dict",
        "Set": "Set",
        "Optional": "Nothing",
    }

    def __init__(self):
        super().__init__()
        self._default_type = "var"
        self._container_type_map = self.CONTAINER_TYPE_MAP

    def usings(self):
        usings = sorted(list(set(self._usings)))
        uses = "\n".join(f"import {mod}" for mod in usings)
        return uses

    def visit_FunctionDef(self, node):
        body = "\n".join([self.visit(n) for n in node.body])
        typenames, args = self.visit(node.args)

        args_list = []
        if len(args) and hasattr(node, "self_type"):
            # implicit this
            del typenames[0]
            del args[0]

        typedecls = []
        index = 0
        for i in range(len(args)):
            typename = typenames[i]
            arg = args[i]
            if typename == "T":
                typename = "T{0}".format(index)
                typedecls.append(typename)
                index += 1
            args_list.append(f"{arg}: {typename}")

        return_type = ""
        if not is_void_function(node):
            if node.returns:
                typename = self._typename_from_annotation(node, attr="returns")
                return_type = f": {typename}"
            else:
                return_type = ": RT"
                typedecls.append("RT")

        template = ""
        if len(typedecls) > 0:
            template = "<{0}>".format(", ".join(typedecls))

        args = ", ".join(args_list)
        funcdef = f"fun {node.name}{template}({args}){return_type} {{"
        return funcdef + "\n" + body + "}\n\n"

    def visit_Return(self, node):
        if node.value:
            ret = self.visit(node.value)
            fndef = None
            for scope in node.scopes:
                if isinstance(scope, ast.FunctionDef):
                    fndef = scope
                    break
            if fndef:
                return_type = self._typename_from_annotation(fndef, attr="returns")
                value_type = get_inferred_kotlin_type(node.value)
                if return_type != value_type and value_type is not None:
                    return f"return {ret} as {return_type}"
            return f"return {ret}"
        return "return"

    def visit_arg(self, node):
        id = get_id(node)
        if id == "self":
            return (None, "self")
        id, _ = self._check_keyword(id)
        typename = "T"
        if node.annotation:
            typename = self._typename_from_annotation(node)
        return (typename, id)

    def visit_Lambda(self, node):
        _, args = self.visit(node.args)
        args_string = ", ".join(args)
        body = self.visit(node.body)
        return f"{{ {args_string} -> {body} }}"

    def visit_Attribute(self, node):
        attr = node.attr

        value_id = self.visit(node.value)

        if not value_id:
            value_id = ""

        if is_class_or_module(value_id, node.scopes):
            return f"{value_id}.{attr}"

        if is_self_arg(value_id, node.scopes):
            return attr

        return f"{value_id}.{attr}"

    def visit_range(self, node, vargs: List[str]) -> str:
        if len(node.args) == 1:
            return "(0..{}-1)".format(vargs[0])
        elif len(node.args) == 2:
            return "({}..{}-1)".format(vargs[0], vargs[1])
        elif len(node.args) == 3:
            return "({}..{}-1 step {})".format(vargs[0], vargs[1], vargs[2])

        raise Exception(
            "encountered range() call with unknown parameters: range({})".format(vargs)
        )

    def visit_print(self, node, vargs: List[str]) -> str:
        def _format(arg):
            if arg.isdigit():
                return arg
            if re.match(r"'.*'", arg) or re.match(r'".*"', arg):
                return arg[1:-1]
            else:
                return f"${arg}"

        vargs_str = " ".join([f"{_format(arg)}" for arg in vargs])
        return f'println("{vargs_str}")'

    def _dispatch(self, node, fname: str, vargs: List[str]) -> Optional[str]:
        dispatch_map = {
            "range": self.visit_range,
            "xrange": self.visit_range,
            "print": self.visit_print,
        }

        if fname in dispatch_map:
            return dispatch_map[fname](node, vargs)

        def visit_min_max(is_max: bool) -> str:
            min_max = "max" if is_max else "min"
            self._usings.add(f"kotlin.math.{min_max}")
            self._typename_from_annotation(node.args[0])
            if hasattr(node.args[0], "container_type"):
                return f"maxOf({vargs[0]})"
            else:
                all_vargs = ", ".join(vargs)
                return f"{min_max}({all_vargs})"

        def visit_floor():
            self._usings.add("kotlin.math.floor")
            return f"floor({vargs[0]}).toInt()"

        # small one liners are inlined here as lambdas
        small_dispatch_map = {
            "int": lambda: f"{vargs[0]}.toInt()",
            "str": lambda: f"{vargs[0]}.toString()",
            # TODO: strings use .length
            "len": lambda: f"{vargs[0]}.size",
            "max": functools.partial(visit_min_max, is_max=True),
            "min": functools.partial(visit_min_max, is_min=True),
            "floor": visit_floor,
            "reversed": lambda: f"{vargs[0]}.reversed()",
        }
        if fname in small_dispatch_map:
            return small_dispatch_map[fname]()
        return None

    def visit_Call(self, node):
        fname = self.visit(node.func)
        vargs = []

        if (
            hasattr(node.func, "value")
            and is_list(node.func.value)
            and fname.endswith(".append")
        ):
            fname = fname.replace(".append", " +=")

        if node.args:
            vargs += [self.visit(a) for a in node.args]
        if node.keywords:
            vargs += [self.visit(kw.value) for kw in node.keywords]

        ret = self._dispatch(node, fname, vargs)
        if ret is not None:
            return ret
        if vargs:
            args = ", ".join(vargs)
        else:
            args = ""

        if fname.endswith("+="):
            return f"{fname} {args}"

        return f"{fname}({args})"

    def visit_For(self, node):
        target = self.visit(node.target)
        it = self.visit(node.iter)
        buf = []
        buf.append("for ({0} in {1}) {{".format(target, it))
        buf.extend([self.visit(c) for c in node.body])
        buf.append("}")
        return "\n".join(buf)

    def visit_Str(self, node):
        return "" + super().visit_Str(node) + ""

    def visit_Bytes(self, node):
        bytes_str = "{0}".format(node.s)
        return bytes_str.replace("'", '"')  # replace single quote with double quote

    def visit_Name(self, node):
        if node.id == "None":
            return "None"
        else:
            return super().visit_Name(node)

    def visit_NameConstant(self, node):
        if node.value is None:
            return "null"
        else:
            return super().visit_NameConstant(node)

    def _make_block(self, node):
        buf = []
        buf.append("if (true) {")
        buf.extend([self.visit(child) for child in node.body])
        buf.append("}")
        return "\n".join(buf)

    def visit_If(self, node):
        body_vars = set([get_id(v) for v in node.scopes[-1].body_vars])
        orelse_vars = set([get_id(v) for v in node.scopes[-1].orelse_vars])
        node.common_vars = body_vars.intersection(orelse_vars)
        return super().visit_If(node)

    def visit_UnaryOp(self, node):
        if isinstance(node.op, ast.USub):
            if isinstance(node.operand, (ast.Call, ast.Num)):
                # Shortcut if parenthesis are not needed
                return "-{0}".format(self.visit(node.operand))
            else:
                return "-({0})".format(self.visit(node.operand))
        else:
            return super().visit_UnaryOp(node)

    def visit_BinOp(self, node):
        if (
            isinstance(node.left, ast.List)
            and isinstance(node.op, ast.Mult)
            and isinstance(node.right, ast.Num)
        ):
            num = self.visit(node.right)
            elt = self.visit(node.left.elts[0])
            return f"Array({num}) {{ {elt} }}"
        else:
            return super().visit_BinOp(node)

    def visit_ClassDef(self, node):
        extractor = DeclarationExtractor(KotlinTranspiler())
        extractor.visit(node)
        declarations = node.declarations = extractor.get_declarations()
        node.class_assignments = extractor.class_assignments
        ret = super().visit_ClassDef(node)
        if ret is not None:
            return ret

        fields = []
        index = 0
        for declaration, typename in declarations.items():
            if typename == None:
                typename = "ST{0}".format(index)
                index += 1
            mut = is_mutable(node.scopes, get_id(declaration))
            mut = "var" if mut else "val"
            fields.append(f"{mut} {declaration}: {typename}")

        for b in node.body:
            if isinstance(b, ast.FunctionDef):
                b.self_type = node.name

        if node.is_dataclass:
            fields = ", ".join(fields)
            body = [self.visit(b) for b in node.body]
            body = "\n".join(body)
            return f"data class {node.name}({fields}) {{\n{body}\n}}\n"
        else:
            fields = "\n".join(fields)
            body = [self.visit(b) for b in node.body]
            body = "\n".join(body)
            return f"class {node.name} {{\n{fields}\n\n {body}\n}}\n"

    def _visit_enum(self, node, typename: str, fields: List[Tuple]):
        fields_list = []

        for field, value in fields:
            fields_list += [
                f"""\
                {field}({value}),
            """
            ]
        fields_str = "".join(fields_list)
        return f"enum class {node.name}(val value: {typename}) {{\n{fields_str}\n}}"

    def visit_StrEnum(self, node):
        fields = []
        for i, (member, var) in enumerate(node.class_assignments.items()):
            var = self.visit(var)
            if var == "auto()":
                var = f'"{member}"'
            fields.append((member, var))
        return self._visit_enum(node, "String", fields)

    def visit_IntEnum(self, node):
        fields = []
        for i, (member, var) in enumerate(node.class_assignments.items()):
            var = self.visit(var)
            if var == "auto()":
                var = i
            fields.append((member, var))
        return self._visit_enum(node, "Int", fields)

    def visit_IntFlag(self, node):
        fields = []
        for i, (member, var) in enumerate(node.class_assignments.items()):
            var = self.visit(var)
            if var == "auto()":
                var = 1 << i
            fields.append((member, var))
        return self._visit_enum(node, "Int", fields)

    def visit_alias(self, node):
        return "use {0}".format(node.name)

    def visit_Import(self, node):
        imports = [self.visit(n) for n in node.names]
        return "\n".join(i for i in imports if i)

    def visit_ImportFrom(self, node):
        if node.module in self.IGNORED_MODULE_LIST:
            return ""

        names = [n.name for n in node.names]
        names = ", ".join(names)
        module_path = node.module.replace(".", "::")
        return "use {0}::{{{1}}}".format(module_path, names)

    def visit_List(self, node):
        elements = [self.visit(e) for e in node.elts]
        elements_str = ", ".join(elements)
        return f"arrayOf({elements_str})"

    def visit_Set(self, node):
        elements = [self.visit(e) for e in node.elts]
        elements_str = ", ".join(elements)
        return f"setOf({elements_str})"

    def visit_Dict(self, node):
        keys = [self.visit(k) for k in node.keys]
        values = [self.visit(k) for k in node.values]
        kv_pairs = ", ".join([f"{k} to {v}" for k, v in zip(keys, values)])
        return f"hashMapOf({kv_pairs})"

    def visit_Subscript(self, node):
        value = self.visit(node.value)
        index = self.visit(node.slice)
        if hasattr(node, "is_annotation"):
            if value in self.CONTAINER_TYPE_MAP:
                value = self.CONTAINER_TYPE_MAP[value]
            if value == "Tuple":
                return "({0})".format(index)
            return "{0}<{1}>".format(value, index)
        return "{0}[{1}]".format(value, index)

    def visit_Index(self, node):
        return self.visit(node.value)

    def visit_Slice(self, node):
        lower = ""
        if node.lower:
            lower = self.visit(node.lower)
        upper = ""
        if node.upper:
            upper = self.visit(node.upper)

        return "{0}..{1}".format(lower, upper)

    def visit_Tuple(self, node):
        elts = [self.visit(e) for e in node.elts]
        elts = ", ".join(elts)
        if hasattr(node, "is_annotation"):
            return elts
        return "({0})".format(elts)

    def visit_unsupported_body(self, name, body):
        buf = ["let {0} = {{ //unsupported".format(name)]
        buf += [self.visit(n) for n in body]
        buf.append("};")
        return buf

    def visit_Try(self, node, finallybody=None):
        buf = self.visit_unsupported_body("try_dummy", node.body)

        for handler in node.handlers:
            buf += self.visit(handler)
        # buf.append("\n".join(excepts));

        if finallybody:
            buf += self.visit_unsupported_body("finally_dummy", finallybody)

        return "\n".join(buf)

    def visit_ExceptHandler(self, node):
        exception_type = ""
        if node.type:
            exception_type = self.visit(node.type)
        name = "except!({0})".format(exception_type)
        body = self.visit_unsupported_body(name, node.body)
        return body

    def visit_Assert(self, node):
        condition = self.visit(node.test)
        return f"assert({condition})"

    def visit_AnnAssign(self, node):
        target = self.visit(node.target)
        type_str = self._typename_from_annotation(node)
        val = self.visit(node.value) if node.value is not None else None
        if type_str == self._default_type:
            return f"var {target} = {val}"
        return f"var {target}: {type_str} = {val}"

    def _visit_AssignOne(self, node, target):
        kw = "var" if is_mutable(node.scopes, get_id(target)) else "val"

        if isinstance(target, ast.Tuple):
            elts = [self.visit(e) for e in target.elts]
            elts_str = ", ".join(elts)
            value = self.visit(node.value)
            if isinstance(node.value, ast.Tuple):
                value = f"Pair{value}"
            return f"{kw} ({elts_str}) = {value}"

        if isinstance(node.scopes[-1], ast.If):
            outer_if = node.scopes[-1]
            target_id = self.visit(target)
            if target_id in outer_if.common_vars:
                value = self.visit(node.value)
                return f"{kw} {target_id} = {value}"

        if isinstance(target, ast.Subscript) or isinstance(target, ast.Attribute):
            target = self.visit(target)
            value = self.visit(node.value)
            return f"{target} = {value}"

        definition = node.scopes.find(target.id)
        if isinstance(target, ast.Name) and defined_before(definition, node):
            target = self.visit(target)
            value = self.visit(node.value)
            return f"{target} = {value}"
        elif isinstance(node.value, ast.List):
            elements = [self.visit(e) for e in node.value.elts]
            elements = ", ".join(elements)
            target = self.visit(target)

            return f"{kw} {target} = arrayOf({elements})"
        else:
            target = self.visit(target)
            value = self.visit(node.value)

            return f"{kw} {target} = {value}"

    def visit_Delete(self, node):
        target = node.targets[0]
        return "{0}.drop()".format(self.visit(target))

    def visit_Raise(self, node):
        if node.exc is not None:
            exc = self.visit(node.exc)
            return f"throw Exception({exc})"
        return "throw Exception()"

    def visit_With(self, node):
        buf = []

        with_statement = "// with!("
        for i in node.items:
            if i.optional_vars:
                with_statement += "{0} as {1}, ".format(
                    self.visit(i.context_expr), self.visit(i.optional_vars)
                )
            else:
                with_statement += "{0}, ".format(self.visit(i.context_expr))
        with_statement = with_statement[:-2] + ") //unsupported\n{"
        buf.append(with_statement)

        for n in node.body:
            buf.append(self.visit(n))

            buf.append("}")

        return "\n".join(buf)

    def visit_Await(self, node):
        expr = self.visit(node.value)
        return f"{expr}.await()"

    def visit_AsyncFunctionDef(self, node):
        fn = self.visit_FunctionDef(node)
        return f"suspend {fn}"

    def visit_Yield(self, node):
        return "//yield is unimplemented"

    def visit_Print(self, node):
        vargs_str = " ".join([f"${arg}" for arg in node.values])
        return f'println("{vargs_str}")'

    def visit_DictComp(self, node):
        return "DictComp /*unimplemented()*/"

    def visit_GeneratorExp(self, node):
        return "GeneratorExp /*unimplemented()*/"

    def visit_ListComp(self, node):
        return self.visit_GeneratorExp(node)  # right now they are the same

    def visit_Global(self, node):
        return "//global {0}".format(", ".join(node.names))

    def visit_Starred(self, node):
        return "starred!({0})/*unsupported*/".format(self.visit(node.value))

    def visit_IfExp(self, node):
        body = self.visit(node.body)
        orelse = self.visit(node.orelse)
        test = self.visit(node.test)
        return f"if ({test}) {body} else {orelse}"



import ast

from py2many.context import add_variable_context, add_list_calls
from py2many.scope import add_scope_context
from py2many.tracer import is_list, is_recursive, value_expr, value_type


def parse(*args):
    source = ast.parse("\n".join(args))
    add_variable_context(source)
    add_scope_context(source)
    return source


class TestValueType:
    def test_direct_assignment(self):
        source = parse("x = 3")
        x = source.body[0]
        t = value_type(x)
        assert t == "3"

    def test_assign_name(self):
        source = parse("x = 3", "y = x")
        y = source.body[1]
        t = value_type(y)
        assert t == "x"

    def test_assign_function(self):
        source = parse("x = 3", "y = foo(x)")
        y = source.body[1]
        t = value_type(y)
        assert t == "foo(x)"

    def test_list_assignment_with_default_values(self):
        source = parse("x = 3", "results = [x]")
        results = source.body[1]
        t = value_type(results)
        assert t == "x"

    def test_list_assignment_with_function_call_as_value(self):
        source = parse("x = 3", "results = [foo(x)]")
        results = source.body[1]
        t = value_type(results)
        assert t == "foo(x)"

    def test_list_assignment_based_on_later_append(self):
        source = parse("x = 3", "results = []", "results.append(x)")
        add_list_calls(source)
        results = source.body[1]
        t = value_type(results)
        assert t == "3"

    def test_list_assignment_with_append_unknown_value(self):
        source = parse("results = []", "x = 3", "results.append(x)")
        add_list_calls(source)
        results = source.body[0]
        t = value_type(results)
        assert t == "3"

    def test_global_list_assignment_with_later_append(self):
        source = parse("results = []", "def add_x():", "   results.append(2)")
        add_list_calls(source)
        results = source.body[0]
        t = value_type(results)
        assert t == "2"


class TestValueExpr:
    def test_catch_long_expression_chain(self):
        source = parse("x = 3 * 1", "y = x - 5", "z = y + 2")
        z = source.body[2]
        t = value_expr(z)
        assert t == "3 * 1 - 5 + 2"

    def test_catch_expression_chain_with_functions(self):
        source = parse("x = 3 * 1", "y = foo(x)", "z = y + 2")
        z = source.body[2]
        t = value_expr(z)
        assert t == "foo(3 * 1) + 2"


# def test_decltype_normal_var():
#     source = parse(
#         "x = 3",
#         "y = foo(x)",
#     )
#     y = source.body[1]
#     t = decltype(y)
#     assert t == "decltype(foo(x))"
#
#
# def test_decltype_list_var():
#     source = parse(
#         "results = []",
#         "x = 3",
#         "results.append(x)",
#     )
#     add_list_calls(source)
#     results = source.body[0]
#     t = decltype(results)
#     assert t == "std::vector<decltype(3)>"


def test_is_list():
    source = parse("list1 = []", "list2 = list1")
    add_list_calls(source)
    list2 = source.body[1]
    assert is_list(list2)


def test_is_recursive():
    source = parse("def rec(n):", "   return rec(n-1) + rec(n)")
    fun = source.body[0]
    assert is_recursive(fun)



import ast


def add_annotation_flags(node):
    return AnnotationTransformer().visit(node)


class AnnotationTransformer(ast.NodeTransformer):
    """
    Adds a flag for every type annotation and nested types so they can be differentiated from array
    """

    def __init__(self):
        self.handling_annotation = False

    def visit_arg(self, node):
        if node.annotation:
            self.handling_annotation = True
            self.visit(node.annotation)
            self.handling_annotation = False
        return node

    def visit_FunctionDef(self, node):
        if node.returns:
            self.handling_annotation = True
            self.visit(node.returns)
            self.handling_annotation = False
        self.generic_visit(node)
        return node

    # without this Dict[x,y] will be translated to HashMap<(x,y)>
    def visit_Tuple(self, node):
        if self.handling_annotation:
            node.is_annotation = True
        self.generic_visit(node)
        return node

    def visit_List(self, node):
        if self.handling_annotation:
            node.is_annotation = True
        self.generic_visit(node)
        return node

    def visit_Name(self, node):
        if self.handling_annotation:
            node.is_annotation = True
        self.generic_visit(node)
        return node

    def visit_Subscript(self, node):
        if self.handling_annotation:
            node.is_annotation = True
        self.generic_visit(node)
        return node

    def visit_AnnAssign(self, node):
        self.handling_annotation = True
        self.visit(node.target)
        self.handling_annotation = False
        self.generic_visit(node)
        return node



import argparse
import ast
import os
import pathlib

from dataclasses import dataclass, field
from distutils import spawn
from subprocess import run
from typing import Callable, List, Optional

from .analysis import add_imports
from .clike import CLikeTranspiler
from .scope import add_scope_context
from .annotation_transformer import add_annotation_flags
from .mutability_transformer import detect_mutable_vars
from .nesting_transformer import detect_nesting_levels
from .context import add_variable_context, add_list_calls
from .inference import infer_types

from py14.transpiler import CppTranspiler, CppListComparisonRewriter
from pyrs.inference import infer_rust_types
from pyrs.transpiler import (
    RustTranspiler,
    RustLoopIndexRewriter,
    RustNoneCompareRewriter,
    RustStringJoinRewriter,
)
from pyjl.transpiler import JuliaTranspiler, JuliaMethodCallRewriter
from pykt.inference import infer_kotlin_types
from pykt.transpiler import KotlinTranspiler, KotlinPrintRewriter
from pynim.inference import infer_nim_types
from pynim.transpiler import NimTranspiler, NimNoneCompareRewriter
from pydart.transpiler import DartTranspiler, DartIntegerDivRewriter
from pygo.inference import infer_go_types
from pygo.transpiler import (
    GoTranspiler,
    GoMethodCallRewriter,
    GoNoneCompareRewriter,
    GoPropagateTypeAnnotation,
    GoVisibilityRewriter,
    GoIfExpRewriter,
)

from py2many.rewriters import (
    ComplexDestructuringRewriter,
    FStringJoinRewriter,
    PythonMainRewriter,
    DocStringToCommentRewriter,
    PrintBoolRewriter,
)

PY2MANY_DIR = pathlib.Path(__file__).parent
ROOT_DIR = PY2MANY_DIR.parent


def core_transformers(tree):
    add_variable_context(tree)
    add_scope_context(tree)
    add_list_calls(tree)
    detect_mutable_vars(tree)
    detect_nesting_levels(tree)
    add_annotation_flags(tree)
    infer_meta = infer_types(tree)
    add_imports(tree)
    return tree, infer_meta


def transpile(source, transpiler, rewriters, transformers, post_rewriters):
    """
    Transpile a single python translation unit (a python script) into
    Rust code.
    """
    if isinstance(source, ast.Module):
        tree = source
    else:
        tree = ast.parse(source)
    language = transpiler.NAME
    generic_rewriters = [
        ComplexDestructuringRewriter(language),
        PythonMainRewriter(language),
        FStringJoinRewriter(language),
        DocStringToCommentRewriter(language),
    ]
    # This is very basic and needs to be run before and after
    # rewrites. Revisit if running it twice becomes a perf issue
    add_scope_context(tree)
    # First run Language independent rewriters
    for rewriter in generic_rewriters:
        tree = rewriter.visit(tree)
    # Language specific rewriters
    for rewriter in rewriters:
        tree = rewriter.visit(tree)
    # Language independent core transformers
    tree, infer_meta = core_transformers(tree)
    # Language specific transformers
    for tx in transformers:
        tx(tree)
    # Language independent rewriters that run after type inference
    generic_post_rewriters = [PrintBoolRewriter(language)]
    for rewriter in generic_post_rewriters:
        tree = rewriter.visit(tree)
    # Language specific rewriters that depend on previous steps
    for rewriter in post_rewriters:
        tree = rewriter.visit(tree)
    # Rerun core transformers
    tree, infer_meta = core_transformers(tree)
    out = []
    code = transpiler.visit(tree) + "\n"
    headers = transpiler.headers(infer_meta)
    features = transpiler.features()
    if features:
        out.append(features)
    if headers:
        out.append(headers)
    usings = transpiler.usings()
    if usings:
        out.append(usings)
    out.append(code)
    return "\n".join(out)


def _create_cmd(parts, filename):
    cmd = [arg.format(filename=filename) for arg in parts]
    if cmd != parts:
        return cmd
    return [*parts, str(filename)]


@dataclass
class LanguageSettings:
    transpiler: CLikeTranspiler
    ext: str
    formatter: Optional[List[str]] = None
    indent: Optional[int] = None
    rewriters: List[ast.NodeVisitor] = field(default_factory=list)
    transformers: List[Callable] = field(default_factory=list)
    post_rewriters: List[ast.NodeVisitor] = field(default_factory=list)
    linter: Optional[List[str]] = None


def cpp_settings(args, env=os.environ):
    clang_format_style = env.get("CLANG_FORMAT_STYLE")
    cxx = env.get("CXX")
    default_cxx = ["clang++", "g++-11", "g++"]
    if cxx:
        if not spawn.find_executable(cxx):
            print(f"Warning: CXX({cxx}) not found")
            cxx = None
    if not cxx:
        for exe in default_cxx:
            if spawn.find_executable(exe):
                cxx = exe
                break
        else:
            cxx = default_cxx[0]
    cxx_flags = env.get("CXXFLAGS")
    if cxx_flags:
        cxx_flags = cxx_flags.split()
    else:
        cxx_flags = ["-std=c++14", "-Wall", "-Werror"]
    cxx_flags = ["-I", str(ROOT_DIR)] + cxx_flags
    if cxx == "clang++":
        cxx_flags += ["-stdlib=libc++"]

    if clang_format_style:
        clang_format_cmd = ["clang-format", f"-style={clang_format_style}", "-i"]
    else:
        clang_format_cmd = ["clang-format", "-i"]

    return LanguageSettings(
        CppTranspiler(),
        ".cpp",
        clang_format_cmd,
        None,
        [CppListComparisonRewriter()],
        linter=[cxx, *cxx_flags],
    )


def rust_settings(args, env=os.environ):
    return LanguageSettings(
        RustTranspiler(),
        ".rs",
        ["rustfmt", "--edition=2018"],
        None,
        [RustNoneCompareRewriter()],
        [infer_rust_types],
        [RustLoopIndexRewriter(), RustStringJoinRewriter()],
    )


def julia_settings(args, env=os.environ):
    format_jl = spawn.find_executable("format.jl")
    if format_jl:
        format_jl = ["julia", "-O0", "--compile=min", "--startup=no", format_jl, "-v"]
    else:
        format_jl = ["format.jl", "-v"]
    return LanguageSettings(
        JuliaTranspiler(), ".jl", format_jl, None, [], [], [JuliaMethodCallRewriter()]
    )


def kotlin_settings(args, env=os.environ):
    return LanguageSettings(
        KotlinTranspiler(),
        ".kt",
        ["ktlint", "-F"],
        transformers=[infer_kotlin_types],
        post_rewriters=[KotlinPrintRewriter()],
        linter=["ktlint"],
    )


def nim_settings(args, env=os.environ):
    nim_args = {}
    nimpretty_args = []
    if args.indent is not None:
        nim_args["indent"] = args.indent
        nimpretty_args.append(f"--indent:{args.indent}")
    return LanguageSettings(
        NimTranspiler(**nim_args),
        ".nim",
        ["nimpretty", *nimpretty_args],
        None,
        [NimNoneCompareRewriter()],
        [infer_nim_types],
    )


def dart_settings(args, env=os.environ):
    return LanguageSettings(
        DartTranspiler(),
        ".dart",
        ["dart", "format"],
        post_rewriters=[DartIntegerDivRewriter()],
    )


def go_settings(args, env=os.environ):
    return LanguageSettings(
        GoTranspiler(),
        ".go",
        ["gofmt", "-w"],
        None,
        [GoNoneCompareRewriter(), GoVisibilityRewriter(), GoIfExpRewriter()],
        [infer_go_types],
        [GoMethodCallRewriter(), GoPropagateTypeAnnotation()],
        linter=["golint", "-set_exit_status", "-min_confidence", "1.0"],
    )


def _get_all_settings(args, env=os.environ):
    return {
        "cpp": cpp_settings(args, env=env),
        "rust": rust_settings(args, env=env),
        "julia": julia_settings(args, env=env),
        "kotlin": kotlin_settings(args, env=env),
        "nim": nim_settings(args, env=env),
        "dart": dart_settings(args, env=env),
        "go": go_settings(args, env=env),
    }


def _process_once(settings, filename, outdir, env=None):
    """Transpile and reformat.

    Returns False if reformatter failed.
    """
    output_path = outdir / (filename.stem + settings.ext)
    if settings.ext == ".kt" and output_path.is_absolute():
        # KtLint does not support absolute path in globs
        output_path = output_path.relative_to(pathlib.Path.cwd())
    print(f"{filename}...{output_path}")
    with open(filename) as f:
        source_data = f.read()
    with open(output_path, "w") as f:
        f.write(
            transpile(
                source_data,
                settings.transpiler,
                settings.rewriters,
                settings.transformers,
                settings.post_rewriters,
            )
        )

    if settings.formatter:
        cmd = _create_cmd(settings.formatter, output_path)
        proc = run(cmd, env=env)
        if proc.returncode:
            print(
                f"Error: {cmd} (code: {proc.returncode}):\n{proc.stderr}{proc.stdout}"
            )
            # format.jl exit code is unreliable
            if settings.ext == ".jl":
                return True
            return False
        if settings.ext == ".kt":
            # ktlint formatter needs to be invoked twice before output is lint free
            if run(cmd, env=env).returncode:
                print(f"Error: Could not reformat: {cmd}")
                return False
    return True


def main(args=None, env=None):
    parser = argparse.ArgumentParser()
    parser.add_argument("--cpp", type=bool, default=False, help="Generate C++ code")
    parser.add_argument("--rust", type=bool, default=False, help="Generate Rust code")
    parser.add_argument("--julia", type=bool, default=False, help="Generate Julia code")
    parser.add_argument(
        "--kotlin", type=bool, default=False, help="Generate Kotlin code"
    )
    parser.add_argument("--nim", type=bool, default=False, help="Generate Nim code")
    parser.add_argument("--dart", type=bool, default=False, help="Generate Dart code")
    parser.add_argument("--go", type=bool, default=False, help="Generate Go code")
    parser.add_argument("--outdir", default=None, help="Output directory")
    parser.add_argument(
        "-i",
        "--indent",
        type=int,
        default=None,
        help="Indentation to use in languages that care",
    )
    args, rest = parser.parse_known_args(args=args)
    for filename in rest:
        settings = cpp_settings(args, env=env)
        if args.cpp:
            pass
        if args.rust:
            settings = rust_settings(args, env=env)
        elif args.julia:
            settings = julia_settings(args, env=env)
        elif args.kotlin:
            settings = kotlin_settings(args, env=env)
        elif args.nim:
            settings = nim_settings(args, env=env)
        elif args.dart:
            settings = dart_settings(args, env=env)
        elif args.go:
            settings = go_settings(args, env=env)
        source = pathlib.Path(filename)
        if args.outdir is None:
            outdir = source.parent
        else:
            outdir = pathlib.Path(args.outdir)

        if source.is_file():
            print(f"Writing to: {outdir}")
            rv = _process_once(settings, source, outdir, env=env)
        else:
            if args.outdir is None:
                outdir = source.parent / f"{source.name}-py2many"

            print(f"Transpiling whole directory to {outdir}:")
            successful = failures = format_errors = 0
            for path in source.rglob("*.py"):
                if path.suffix != ".py":
                    continue
                if path.parent.name == "__pycache__":
                    continue

                relative_path = path.relative_to(source)
                target_path = outdir / relative_path
                target_dir = target_path.parent
                os.makedirs(target_dir, exist_ok=True)

                try:
                    if _process_once(settings, path, target_dir, env=env):
                        print(f"Error: Could not reformat: {path}")
                        format_errors += 1
                    successful += 1
                except Exception as e:
                    failures += 1
                    print(f"Error: Could not transpile: {path}")
                    print(f"Due to: {e}")

            print("\nFinished!")
            print(f"Successful: {successful}")
            if format_errors:
                print(f"Failed to reformat: {format_errors}")
            print(f"Failed to convert: {failures}")
            print()
            rv = not failures or format_errors
        return rv



import ast
from .scope import ScopeMixin


def add_list_calls(node):
    """Provide context to Module and Function Def"""
    return ListCallTransformer().visit(node)


def add_variable_context(node):
    """Provide context to Module and Function Def"""
    return VariableTransformer().visit(node)


class ListCallTransformer(ast.NodeTransformer):
    """
    Adds all calls to list to scope block.
    You need to apply VariableTransformer before you use it.
    """

    def visit_Call(self, node):
        if self.is_list_addition(node):
            var = node.scopes.find(node.func.value.id)
            if self.is_list_assignment(var.assigned_from):
                if not hasattr(var, "calls"):
                    var.calls = []
                var.calls.append(node)
        return node

    def is_list_assignment(self, node):
        return (
            hasattr(node, "value")
            and isinstance(node.value, ast.List)
            and hasattr(node, "targets")
            and isinstance(node.targets[0].ctx, ast.Store)
        )

    def is_list_addition(self, node):
        """Check if operation is adding something to a list"""
        list_operations = ["append", "extend", "insert"]
        return (
            isinstance(node.func.ctx, ast.Load)
            and hasattr(node.func, "value")
            and isinstance(node.func.value, ast.Name)
            and node.func.attr in list_operations
        )


class VariableTransformer(ast.NodeTransformer, ScopeMixin):
    """Adds all defined variables to scope block"""

    def visit_FunctionDef(self, node):
        node.vars = []
        # So function signatures are accessible even after they're
        # popped from the scope
        self.scopes[-2].vars.append(node)
        for arg in node.args.args:
            arg.assigned_from = node
            node.vars.append(arg)
        self.generic_visit(node)
        return node

    def visit_ClassDef(self, node):
        node.vars = []
        # So classes are accessible even after they're
        # popped from the scope
        self.scopes[-2].vars.append(node)
        self.generic_visit(node)
        return node

    def visit_Import(self, node):
        for name in node.names:
            name.imported_from = node

    def visit_If(self, node):
        node.vars = []
        self.visit(node.test)
        for e in node.body:
            self.visit(e)
        node.body_vars = node.vars

        node.vars = []
        for e in node.orelse:
            self.visit(e)
        node.orelse_vars = node.vars

        node.vars = []
        return node

    def visit_For(self, node):
        node.target.assigned_from = node
        node.vars = [node.target]
        self.generic_visit(node)
        return node

    def visit_Module(self, node):
        node.vars = []
        self.generic_visit(node)
        return node

    def visit_With(self, node):
        node.vars = []
        self.generic_visit(node)
        return node

    def visit(self, node):
        with self.enter_scope(node):
            return super().visit(node)

    def visit_Assign(self, node):
        for target in node.targets:
            if isinstance(target, ast.Name):
                target.assigned_from = node
                self.scope.vars.append(target)
        return node

    def visit_AnnAssign(self, node):
        target = node.target
        if isinstance(target, ast.Name):
            target.assigned_from = node
            self.scope.vars.append(target)
        return node

    def visit_AugAssign(self, node):
        target = node.target
        if isinstance(target, ast.Name):
            target.assigned_from = node
            self.scope.vars.append(target)
        return node



code_0 = 0
code_1 = 1

code_a = "a"
code_b = "b"

l_b = {
    code_a,
}

l_c = {
    code_b: code_0,
}

if __name__ == "__main__":
    assert "a" in l_b
    # Dereferencing items in the dict fails on Rust.
    # Dict compare 'in' does not work on Dart
    # Print of sets and dicts fails on Rust
    print("OK")



from py2many.declaration_extractor import (
    DeclarationExtractor as CommonDeclarationExtractor,
)
from .inference import RUST_TYPE_MAP


class DeclarationExtractor(CommonDeclarationExtractor):
    def __init__(self, transpiler):
        super().__init__(transpiler)
        self._type_map = RUST_TYPE_MAP

    # TODO better type infering based on variable init
    def type_by_initialization(self, init_str):
        if init_str == "vec![]":
            return "Vec<_>"
        elif init_str == "HashMap::new()":
            return "HashMap<_,_>"
        elif init_str == "None":
            return "Option<_>"
        elif init_str == "true" or init_str == "false":
            return "bool"
        else:
            return None



def fib(i: int) -> int:
    if i == 0 or i == 1:
        return 1
    return fib(i - 1) + fib(i - 2)


if __name__ == "__main__":
	# Using a variable to workaround
	# https://github.com/adsharma/py2many/issues/64
	rv = fib(5)
	print(rv)



#!/usr/bin/env python3


def compare_assert(a: int, b: int):
    assert a == b
    assert not (0 == 1)


if __name__ == "__main__":
    assert True
    assert not False
    compare_assert(1, 1)
    # assert not None
    # assert not 0
    # with message
    assert True, 1
    assert True, "message"
    print("OK")



#!/usr/bin/env python3

from typing import List

def bubble_sort(seq: List[int]) -> List[int]:
    L = len(seq)
    for _ in range(L):
        for n in range(1, L):
            if seq[n] < seq[n - 1]:
                seq[n - 1], seq[n] = seq[n], seq[n - 1]
    return seq


if __name__ == "__main__":
    unsorted = [14, 11, 19, 5, 16, 10, 19, 12, 5, 12]
    expected = [5, 5, 10, 11, 12, 12, 14, 16, 19, 19]

    assert bubble_sort(unsorted) == expected
    print("OK")



from py2many.declaration_extractor import (
    DeclarationExtractor as CommonDeclarationExtractor,
)
from .inference import KT_TYPE_MAP


class DeclarationExtractor(CommonDeclarationExtractor):
    def __init__(self, transpiler):
        super().__init__(transpiler)
        self._type_map = KT_TYPE_MAP



import ast


def detect_nesting_levels(node):
    return NestingTransformer().visit(node)


class NestingTransformer(ast.NodeTransformer):
    """
    Some languages are white space sensitive. This transformer
    annotates relevant nodes with the nesting level
    """

    def __init__(self):
        self.level = 0

    def visit_FunctionDef(self, node):
        node.level = self.level
        self.level += 1
        self.generic_visit(node)
        self.level -= 1
        return node

    def visit_ClassDef(self, node):
        node.level = self.level
        self.level += 1
        self.generic_visit(node)
        self.level -= 1
        return node

    def visit_If(self, node):
        node.level = self.level
        self.level += 1
        self.generic_visit(node)
        self.level -= 1
        return node

    def visit_While(self, node):
        node.level = self.level
        self.level += 1
        self.generic_visit(node)
        self.level -= 1
        return node

    def visit_For(self, node):
        node.level = self.level
        self.level += 1
        self.generic_visit(node)
        self.level -= 1
        return node

    def visit_Assign(self, node):
        node.level = self.level
        self.generic_visit(node)
        return node



import argparse
import os.path
import unittest
import sys

from distutils import spawn
from pathlib import Path
from subprocess import run
from unittest.mock import Mock
from unittest_expander import foreach, expand

from py2many.cli import main, _create_cmd, _get_all_settings

TESTS_DIR = Path(__file__).parent.absolute()
ROOT_DIR = TESTS_DIR.parent

KEEP_GENERATED = os.environ.get("KEEP_GENERATED", False)
UPDATE_EXPECTED = os.environ.get("UPDATE_EXPECTED", False)
CXX = os.environ.get("CXX", "clang++")
ENV = {
    "cpp": {
        "CLANG_FORMAT_STYLE": "LLVM",
    },
    "rust": {
        "RUSTFLAGS": "--deny warnings",
    },
}
COMPILERS = {
    "cpp": [CXX, "-std=c++14", "-I", str(ROOT_DIR)]
    + (["-stdlib=libc++"] if CXX == "clang++" else []),
    "dart": ["dart", "compile", "exe"],
    "go": ["go", "build"],
    "julia": ["julia", "--compiled-modules=yes"],
    "kotlin": ["kotlinc"],
    "nim": ["nim", "compile", "--nimcache:."],
    "rust": ["cargo", "script", "--build-only", "--debug"],
}
INVOKER = {
    "go": ["go", "run"],
    "julia": ["julia", "--compiled-modules=yes"],
    "kotlin": ["kscript"],
    "rust": ["cargo", "script"],
}

TEST_CASES = [
    item.stem
    for item in (TESTS_DIR / "cases").glob("*.py")
    if not item.stem.startswith("test_")
]

EXPECTED_LINT_FAILURES = ["int_enum.go", "rect.go", "str_enum.go"]

EXPECTED_COMPILE_FAILURES = []


def has_main(filename):
    with open(filename) as f:
        lines = f.readlines()
    return bool(
        [line in line for line in lines if "def main" in line or "__main__" in line]
    )


@expand
class CodeGeneratorTests(unittest.TestCase):
    LANGS = list(_get_all_settings(Mock(indent=4)).keys())
    maxDiff = None

    SHOW_ERRORS = os.environ.get("SHOW_ERRORS", False)
    KEEP_GENERATED = os.environ.get("KEEP_GENERATED", False)
    UPDATE_EXPECTED = os.environ.get("UPDATE_EXPECTED", False)
    LINT = os.environ.get("LINT", True)

    def setUp(self):
        os.chdir(TESTS_DIR)

    @foreach(sorted(LANGS))
    @foreach(sorted(TEST_CASES))
    def test_generated(self, case, lang):
        env = os.environ.copy()
        if ENV.get(lang):
            env.update(ENV.get(lang))

        settings = _get_all_settings(Mock(indent=4), env=env)[lang]
        ext = settings.ext
        if (
            not self.UPDATE_EXPECTED
            and not self.KEEP_GENERATED
            and not os.path.exists(f"expected/{case}{ext}")
        ):
            raise unittest.SkipTest(f"expected/{case}{ext} not found")

        if settings.formatter:
            if not spawn.find_executable(settings.formatter[0]):
                raise unittest.SkipTest(f"{settings.formatter[0]} not available")

        if ext == ".kt":
            class_name = str(case.title()) + "Kt"
            exe = TESTS_DIR / (class_name + ".class")
        elif ext == ".cpp":
            exe = TESTS_DIR / "a.out"
        elif ext == ".dart" or (ext == ".nim" and sys.platform == "win32"):
            exe = TESTS_DIR / "cases" / f"{case}.exe"
        else:
            exe = TESTS_DIR / "cases" / f"{case}"
        exe.unlink(missing_ok=True)

        case_filename = TESTS_DIR / "cases" / f"{case}.py"
        case_output = TESTS_DIR / "cases" / f"{case}{ext}"
        is_script = has_main(case_filename)
        self.assertTrue(is_script)

        proc = run([sys.executable, str(case_filename)], capture_output=True)
        expected_output = proc.stdout
        if proc.returncode:
            raise RuntimeError(
                f"Invalid cases/{case}.py:\n{expected_output}{proc.stderr}"
            )
        self.assertTrue(expected_output, "Test cases must print something")
        expected_output = expected_output.splitlines()

        args = [f"--{lang}=1", str(case_filename)]

        try:
            rv = main(args=args, env=env)
            with open(f"cases/{case}{ext}") as actual:
                generated = actual.read()
                if os.path.exists(f"expected/{case}{ext}") and not self.UPDATE_EXPECTED:
                    with open(f"expected/{case}{ext}") as f2:
                        self.assertEqual(f2.read(), generated)
                        print("expected = generated")

            expect_failure = (
                not self.SHOW_ERRORS and f"{case}{ext}" in EXPECTED_LINT_FAILURES
            )

            if not expect_failure:
                assert rv, "formatting failed"
            elif not rv:
                raise unittest.SkipTest("formatting failed")

            compiler = COMPILERS[lang]
            if compiler:
                if not spawn.find_executable(compiler[0]):
                    raise unittest.SkipTest(f"{compiler[0]} not available")
                expect_compile_failure = (
                    not self.SHOW_ERRORS and f"{case}{ext}" in EXPECTED_COMPILE_FAILURES
                )
                if expect_compile_failure:
                    return
                proc = run(
                    [*compiler, f"cases/{case}{ext}"], env=env, check=not expect_failure
                )

                if proc.returncode:
                    raise unittest.SkipTest(f"{case}{ext} doesnt compile")

                if self.UPDATE_EXPECTED or not os.path.exists(f"expected/{case}{ext}"):
                    with open(f"expected/{case}{ext}", "w") as f:
                        f.write(generated)

            stdout = None
            if exe.exists() and os.access(exe, os.X_OK):
                stdout = run([exe], env=env, capture_output=True, check=True).stdout

            elif INVOKER.get(lang):
                invoker = INVOKER.get(lang)
                if not spawn.find_executable(invoker[0]):
                    raise unittest.SkipTest(f"{invoker[0]} not available")
                proc = run(
                    [*invoker, case_output],
                    env=env,
                    capture_output=True,
                    check=not expect_failure,
                )

                stdout = proc.stdout

                if proc.returncode:
                    raise unittest.SkipTest(f"Execution of {case}{ext} failed")
            else:
                raise RuntimeError("Compiled output not detected")

            if expected_output and stdout:
                stdout = stdout.splitlines()
                self.assertEqual(expected_output, stdout)

                if settings.linter and self.LINT:
                    if not spawn.find_executable(settings.linter[0]):
                        raise unittest.SkipTest(f"{settings.linter[0]} not available")
                    if settings.ext == ".kt" and case_output.is_absolute():
                        # KtLint does not support absolute path in globs
                        case_output = case_output.relative_to(Path.cwd())
                    linter = _create_cmd(settings.linter, case_output)
                    if ext == ".cpp":
                        linter.append("-Wno-unused-variable")
                        if case == "coverage":
                            linter.append(
                                "-Wno-null-arithmetic"
                                if CXX == "clang++"
                                else "-Wno-pointer-arith"
                            )
                    proc = run(linter, env=env)
                    # golint is failing regularly due to exports without docs
                    if proc.returncode and linter[0] == "golint":
                        expect_failure = True
                    if proc.returncode and expect_failure:
                        raise unittest.SkipTest(f"{case}{ext} failed linter")
                    self.assertFalse(proc.returncode)

                    if expect_failure:
                        raise AssertionError(f"{case}{ext} passed unexpectedly")

        finally:
            if not self.KEEP_GENERATED:
                case_output.unlink(missing_ok=True)
            exe.unlink(missing_ok=True)

    @foreach(sorted(TEST_CASES))
    # This test name must be alpha before `test_generated` otherwise
    # KEEP_GENERATED does not work.
    def test_env_cxx_gcc(self, case):
        lang = "cpp"
        ext = ".cpp"
        if not os.path.exists(f"expected/{case}{ext}"):
            raise unittest.SkipTest(f"expected/{case}{ext} not found")

        env = os.environ.copy()
        env["CXX"] = "g++-11" if sys.platform == "darwin" else "g++"
        env["CXXFLAGS"] = "-std=c++14 -Wall -Werror"

        if not spawn.find_executable(env["CXX"]):
            raise unittest.SkipTest(f"{env['CXX']} not available")

        settings = _get_all_settings(Mock(indent=4), env=env)[lang]
        assert settings.linter[0].startswith("g++")

        if not spawn.find_executable("astyle"):
            raise unittest.SkipTest("astyle not available")

        settings.formatter = ["astyle"]

        exe = TESTS_DIR / "a.out"
        exe.unlink(missing_ok=True)

        case_filename = TESTS_DIR / "cases" / f"{case}.py"
        case_output = TESTS_DIR / "cases" / f"{case}{ext}"

        args = [f"--{lang}=1", str(case_filename)]

        linter = _create_cmd(settings.linter, case_output)

        try:
            rv = main(args=args, env=env)
            assert rv

            linter.append("-Wno-unused-variable")
            if case == "coverage":
                linter.append("-Wno-pointer-arith")
            proc = run(linter, env=env)
            assert not proc.returncode
        except FileNotFoundError as e:
            raise unittest.SkipTest(f"Failed invoking {env['CXX']} or {linter}: {e}")
        finally:
            if not KEEP_GENERATED:
                case_output.unlink(missing_ok=True)
            exe.unlink(missing_ok=True)

    def test_env_clang_format_style(self):
        lang = "cpp"
        env = {
            "CLANG_FORMAT_STYLE": "Google",
        }
        settings = _get_all_settings(Mock(indent=4), env=env)[lang]
        self.assertIn("-style=Google", settings.formatter)

    def test_arg_nim_indent(self):
        lang = "nim"
        settings = _get_all_settings(Mock(indent=2))[lang]
        self.assertIn("--indent:2", settings.formatter)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--lint", type=bool, default=False, help="Lint generated code")
    parser.add_argument(
        "--show-errors", type=bool, default=False, help="Show compile errors"
    )
    parser.add_argument(
        "--keep-generated",
        type=bool,
        default=False,
        help="Keep generated code for debug",
    )
    parser.add_argument(
        "--update-expected", type=bool, default=False, help="Update tests/expected"
    )
    args, rest = parser.parse_known_args()

    CodeGeneratorTests.SHOW_ERRORS |= args.show_errors
    CodeGeneratorTests.KEEP_GENERATED |= args.keep_generated
    CodeGeneratorTests.UPDATE_EXPECTED |= args.update_expected
    CodeGeneratorTests.LINT |= args.lint

    rest = [sys.argv[0]] + rest
    unittest.main(argv=rest)



"""This file implements a rectangle class """

from dataclasses import dataclass


@dataclass
class Rectangle:
    """Rectangle as a dataclass"""
    height: int
    length: int

    def is_square(self) -> bool:
        """Go likes this to be camel case"""
        return self.height == self.length


def show():
    r = Rectangle(height=1, length=1)
    assert r.is_square()

    r = Rectangle(height=1, length=2)
    assert not r.is_square()

    print(r.height)
    print(r.length)


if __name__ == "__main__":
    show()



from enum import Enum


class Colors(str, Enum):
    RED = "red"
    GREEN = "green"
    BLUE = "blue"


def show():
    color_map = {Colors.RED: "1", Colors.GREEN: "2", Colors.BLUE: "3"}
    a = Colors.GREEN
    if a == Colors.GREEN:
        print("green")
    else:
        print("Not green")
    print(len(color_map))


if __name__ == "__main__":
    show()



import ast
import os.path
import unittest
import sys
from distutils import spawn
from pathlib import Path
from subprocess import run
from textwrap import dedent
from unittest.mock import Mock

import astpretty

from unittest_expander import foreach, expand

from py2many.cli import transpile, _get_all_settings

TESTS_DIR = Path(__file__).parent
ROOT_DIR = TESTS_DIR.parent

KEEP_GENERATED = os.environ.get("KEEP_GENERATED", False)
UPDATE_EXPECTED = os.environ.get("UPDATE_EXPECTED", False)
CXX = os.environ.get("CXX", "clang++")
ENV = {"rust": {"RUSTFLAGS": "--deny warnings"}}
COMPILERS = {
    "cpp": [CXX, "-std=c++14", "-I", str(ROOT_DIR), "-stdlib=libc++"],
    "dart": ["dart", "compile", "exe"],
    "go": ["go", "build"],
    "julia": ["julia", "--compiled-modules=yes"],
    "kotlin": ["kotlinc"],
    "nim": ["nim", "compile", "--nimcache:."],
    "rust": ["cargo", "script", "--build-only", "--debug"],
}
INVOKER = {
    "go": ["go", "run"],
    "julia": ["julia", "--compiled-modules=yes"],
    "kotlin": ["kscript"],
    "rust": ["cargo", "script"],
}

_INT_ENUM = dedent(
    """
    from enum import IntEnum, auto
    class Colors(IntEnum):
        RED = auto()
        GREEN = auto()
        BLUE = auto()
"""
)
_INT_FLAG = dedent(
    """
    from enum import IntFlag
    class Permissions(IntFlag):
        R = 1
        W = 2
        X = 16
    a = Permissions.R | Permissions.W
"""
)

TEST_CASES = {
    "assert_0": "assert not 0",
    "assert_None": "assert not None",
    "dict_value_type": "a = {1: 1, 2: 2.0}",  # https://github.com/adsharma/py2many/issues/171
    "empty_print": "print()",
    "fstring": "assert f'{1+1}'",  # https://github.com/adsharma/py2many/issues/74
    "str_format": "ab = '{}{}'.format('a', 'b')",  # https://github.com/adsharma/py2many/issues/73
    "percent_formatting": "a = '~ %s ~' % 'a'",  # https://github.com/adsharma/py2many/issues/176
    "tuple_destruct": "foo, (baz, qux) = 4, (5, 6); assert foo != (baz != qux)",  # https://github.com/adsharma/py2many/issues/155
    "float_1": "a = float(1)",  # https://github.com/adsharma/py2many/issues/129
    "print_None": "print(None)",
    "class_vars": "class A:\n  B = 'FOO'\ndef main(): assert A.B == 'FOO'",  # https://github.com/adsharma/py2many/issues/144
    "default_init": dedent(
        """
        class Rectangle:
            def __init__(self, height=0, length=0):
                self.height = height
                self.length = length

        def main(): r = Rectangle()
    """
    ),  # https://github.com/adsharma/py2many/issues/143
    "complex_dict": "a = ('a', ); b = {a: 'a'}; assert 'a' == b[a]",
    "simple_dict": "l_b = {'a': 0}; assert 'a' in l_b",
    "annassign_List": "from typing import List; items : List = ['a', 'b']; print(items)",
    "intenum_iter": f"{_INT_ENUM}\ndef main():\n  for val in Colors:    print(val)",  # https://github.com/adsharma/py2many/issues/31
    "intflag_bitop": f"{_INT_FLAG}\ndef main():\n  if a & Permissions.R:    print('R')",  # https://github.com/adsharma/py2many/issues/115
    "bool_to_int": "print(int(True))",  # https://github.com/adsharma/py2many/issues/130
    "bool_plus_int": "print(True+1)",  # https://github.com/adsharma/py2many/issues/131
    "del": "a = 1; del a",
    "dict_empty": "assert not {}",
    "dict_del": "a = {1: 1}; del a[1]; assert not a",
    "dict_order": "a = {1: 1, 2: 2, 3: 3}; del a[2]; a[2] = 2; assert list(a.keys()) == [1, 3, 2]",
}

# NOTE: Inclusion here does not indicate that the case is comparable
EXPECTED_SUCCESSES = [
    "annassign_List.dart",
    "assert_None.dart",
    "bool_plus_int.jl",
    "bool_plus_int.rs",
    "bool_to_int.jl",
    "bool_to_int.nim",
    "bool_to_int.rs",
    "complex_dict.jl",
    "complex_dict.nim",
    "del.rs",
    "dict_value_type.dart",
    "dict_value_type.jl",
    "dict_value_type.kt",
    "empty_print.dart",
    "empty_print.jl",
    "empty_print.kt",
    "empty_print.rs",
    "intenum_iter.nim",
    "float_1.jl",
    "float_1.nim",
    "print_None.dart",
    "print_None.jl",
    "simple_dict.kt",
    "simple_dict.nim",
    "str_format.go",
    "str_format.kt",
    "tuple_destruct.jl",
]


def has_main(source):
    lines = source.splitlines()
    return bool(
        [line in line for line in lines if "def main" in line or "__main__" in line]
    )


@expand
class CodeGeneratorTests(unittest.TestCase):
    LANGS = list(_get_all_settings(Mock(indent=4)).keys())
    maxDiff = None

    def setUp(self):
        os.chdir(TESTS_DIR)

    @foreach(sorted(LANGS))
    @foreach(sorted(TEST_CASES.keys()))
    def test_snippet(self, case, lang):
        env = os.environ.copy()
        if ENV.get(lang):
            env.update(ENV.get(lang))

        settings = _get_all_settings(Mock(indent=4), env=env)[lang]
        ext = settings.ext
        source_data = TEST_CASES[case]
        is_script = has_main(source_data)
        if ext in [".dart", ".kt", ".rs"] and not is_script:
            source_data = f"def main():\n  {source_data}"
        print(f">{source_data}<")
        tree = ast.parse(source_data)
        astpretty.pprint(tree)
        proc = run([sys.executable, "-c", source_data], capture_output=True)
        if proc.returncode:
            raise RuntimeError(f"Invalid case {case}:\n{proc.stdout}{proc.stderr}")
        try:
            result = transpile(
                tree,
                settings.transpiler,
                settings.rewriters,
                settings.transformers,
                settings.post_rewriters,
            )
        except NotImplementedError as e:
            raise unittest.SkipTest(str(e))

        if settings.formatter:
            if not spawn.find_executable(settings.formatter[0]):
                raise unittest.SkipTest(f"{settings.formatter[0]} not available")

        if ext == ".kt":
            class_name = str(case.title()) + "Kt"
            exe = TESTS_DIR / (class_name + ".class")
        elif ext == ".cpp":
            exe = TESTS_DIR / "a.out"
        elif ext == ".dart" or (ext == ".nim" and sys.platform == "win32"):
            exe = TESTS_DIR / "cases" / f"{case}.exe"
        else:
            exe = TESTS_DIR / "cases" / f"{case}"
        exe.unlink(missing_ok=True)

        case_output = TESTS_DIR / "cases" / f"{case}{ext}"
        expect_success = f"{case}{ext}" in EXPECTED_SUCCESSES

        with open(case_output, "w") as f:
            f.write(result)

        if settings.formatter:
            if settings.ext == ".kt" and case_output.is_absolute():
                # KtLint does not support absolute path in globs
                case_output = case_output.relative_to(Path.cwd())
            proc = run([*settings.formatter, case_output], env=env, capture_output=True)
            if proc.returncode:
                raise unittest.SkipTest(
                    f"Error: Could not reformat using {settings.formatter}:\n{proc.stdout}{proc.stderr}"
                )

        try:
            compiler = COMPILERS[lang]
            if compiler:
                if not spawn.find_executable(compiler[0]):
                    raise unittest.SkipTest(f"{compiler[0]} not available")
                proc = run([*compiler, case_output], env=env, capture_output=True)

                if proc.returncode and not expect_success:
                    raise unittest.SkipTest(
                        f"{case}{ext} doesnt compile:\n{proc.stdout}{proc.stderr}"
                    )

            if exe.exists() and os.access(exe, os.X_OK):
                if not expect_success:
                    raise AssertionError(f"{case}{ext} compiled")

            elif INVOKER.get(lang):
                invoker = INVOKER.get(lang)
                if not spawn.find_executable(invoker[0]):
                    raise unittest.SkipTest(f"{invoker[0]} not available")
                proc = run([*invoker, case_output], env=env, capture_output=True)

                if proc.returncode and not expect_success:
                    raise unittest.SkipTest(
                        f"Execution of {case}{ext} failed:\n{proc.stdout}{proc.stderr}"
                    )
                if not expect_success:
                    raise AssertionError(f"{case}{ext} invoked")
            else:
                return

        finally:
            if not KEEP_GENERATED:
                case_output.unlink(missing_ok=True)
            exe.unlink(missing_ok=True)

        if not expect_success:
            assert False


if __name__ == "__main__":
    unittest.main()



# please refer to __init__.py file for an explanation


def has_even(numbers):
    for num in numbers:
        if num % 2 == 0:
            return True
    return False


def main():
    vec = [1, 9, 2, 5, 4]
    even_exists = has_even(vec)
    print("Has even number", even_exists)



import ast

from py2many.clike import CLikeTranspiler as CommonCLikeTranspiler

py14_type_map = {
    "bool": "bool",
    "int": "int",
    "float": "double",
    "bytes": "byte[]",
    "str": "std::string",
    "c_int8": "int8_t",
    "c_int16": "int16_t",
    "c_int32": "int32_t",
    "c_int64": "int64_t",
    "c_uint8": "uint8_t",
    "c_uint16": "uint16_t",
    "c_uint32": "uint32_t",
    "c_uint64": "uint64_t",
}


class CLikeTranspiler(CommonCLikeTranspiler):
    def __init__(self):
        super().__init__()
        self._type_map = py14_type_map

    def visit_BinOp(self, node):
        if isinstance(node.op, ast.Pow):
            return "std::pow({0}, {1})".format(
                self.visit(node.left), self.visit(node.right)
            )
        left = self.visit(node.left)
        if not isinstance(node.left, ast.Name) and not isinstance(
            node.left, ast.Constant
        ):
            left = f"({left})"
        right = self.visit(node.right)
        if not isinstance(node.right, ast.Name) and not isinstance(
            node.right, ast.Constant
        ):
            right = f"({right})"
        return " ".join([left, self.visit(node.op), right])

    def visit_In(self, node):
        self._headers.append("#include <algorithm>")
        left = self.visit(node.left)
        right = self.visit(node.comparators[0])
        return f"(std::find({right}.begin(), {right}.end(), {left}) != {right}.end())"



import ast

from py2many.clike import CLikeTranspiler as CommonCLikeTranspiler

from .inference import RUST_WIDTH_RANK, RUST_TYPE_MAP, is_rust_reference


# allowed as names in Python but treated as keywords in Rust
rust_keywords = frozenset(
    [
        "struct",
        "type",
        "match",
        "impl",
        "const",
        "enum",
        "extern",
        "fn",
        "loop",
        "move",
        "mut",
        "pub",
        "ref",
        "trait",
        "where",
        "use",
        "unsafe",
    ]
)


class CLikeTranspiler(CommonCLikeTranspiler):
    def __init__(self):
        super().__init__()
        self._type_map = RUST_TYPE_MAP

    def visit_Name(self, node):
        if node.id in rust_keywords:
            return node.id + "_"
        return super().visit_Name(node)

    def visit_BinOp(self, node):
        if isinstance(node.op, ast.Pow):
            return "pow({0}, {1})".format(self.visit(node.left), self.visit(node.right))

        left = self.visit(node.left)
        op = self.visit(node.op)
        right = self.visit(node.right)

        left_type = self._typename_from_annotation(node.left)
        right_type = self._typename_from_annotation(node.right)

        left_rank = RUST_WIDTH_RANK.get(left_type, -1)
        right_rank = RUST_WIDTH_RANK.get(right_type, -1)

        if left_rank > right_rank:
            right = f"{right} as {left_type}"
        elif right_rank > left_rank:
            left = f"{left} as {right_type}"

        # Multiplication and division binds tighter (has higher precedence) than addition and subtraction.
        # To visually communicate this we omit spaces when multiplying and dividing.
        if isinstance(node.op, (ast.Mult, ast.Div)):
            return f"({left}{op}{right})"
        else:
            return f"({left} {op} {right})"

    def visit_Compare(self, node):
        if isinstance(node.ops[0], ast.In):
            return self.visit_In(node)

        node_right = node.comparators[0]
        left_type = self._typename_from_annotation(node.left)
        right_type = self._typename_from_annotation(node_right)

        left = self.visit(node.left)
        op = self.visit(node.ops[0])
        right = self.visit(node_right)

        if not is_rust_reference(node.left) and is_rust_reference(node_right):
            right = f"*{right}"

        if is_rust_reference(node.left) and not is_rust_reference(node_right):
            left = f"*{left}"

        left_rank = RUST_WIDTH_RANK.get(left_type, -1)
        right_rank = RUST_WIDTH_RANK.get(right_type, -1)

        if left_rank > right_rank:
            right = f"{right} as {left_type}"
        elif right_rank > left_rank:
            left = f"{left} as {right_type}"

        return f"{left} {op} {right}"

    def visit_In(self, node):
        left = self.visit(node.left)
        right = self.visit(node.comparators[0])
        return "{0}.any({1})".format(right, left)



import ast
import functools
import textwrap

from .clike import CLikeTranspiler
from .declaration_extractor import DeclarationExtractor
from py2many.tracer import is_list, defined_before, is_class_or_module, is_self_arg

from py2many.analysis import get_id, is_mutable, is_void_function
from py2many.inference import get_inferred_type

from typing import Optional, List


class DartIntegerDivRewriter(ast.NodeTransformer):
    def visit_BinOp(self, node):
        self.generic_visit(node)
        if isinstance(node.op, ast.Div):
            left_type = get_id(get_inferred_type(node.left))
            right_type = get_id(get_inferred_type(node.right))
            if set([left_type, right_type]) == {"int"}:
                # This attribute should technically be on node.op
                # But python seems to use the same AST node for other
                # division operations?
                node.use_integer_div = True
        return node


class DartTranspiler(CLikeTranspiler):
    NAME = "dart"

    CONTAINER_TYPE_MAP = {
        "List": "List",
        "Dict": "Map",
        "Set": "Set",
        "Optional": "Nothing",
    }

    def __init__(self):
        super().__init__()
        self._container_type_map = self.CONTAINER_TYPE_MAP
        self._default_type = "var"
        self._temp = 0

    def _get_temp(self):
        self._temp += 1
        return f"__tmp{self._temp}"

    def usings(self):
        usings = sorted(list(set(self._usings)))
        uses = "\n".join(f"import '{mod}';" for mod in usings)
        return f"// @dart=2.9\n{uses}" if uses else ""

    def visit_FunctionDef(self, node):
        body = "\n".join([self.visit(n) for n in node.body])
        typenames, args = self.visit(node.args)

        args_list = []
        if len(args) and hasattr(node, "self_type"):
            # implicit this
            del typenames[0]
            del args[0]

        typedecls = []
        index = 0
        for i in range(len(args)):
            typename = typenames[i]
            arg = args[i]
            if typename == "T":
                typename = "T{0}".format(index)
                typedecls.append(typename)
                index += 1
            args_list.append(f"{typename} {arg}")

        return_type = ""
        if not is_void_function(node):
            if node.returns:
                typename = self._typename_from_annotation(node, attr="returns")
                return_type = f"{typename}"
            else:
                return_type = "RT"
                typedecls.append("RT")

        template = ""
        if len(typedecls) > 0:
            template = "<{0}>".format(", ".join(typedecls))

        args = ", ".join(args_list)
        funcdef = f"{return_type} {node.name}{template}({args}) {{"
        return funcdef + "\n" + body + "}\n\n"

    def visit_Return(self, node):
        if node.value:
            return "return {0};".format(self.visit(node.value))
        return "return;"

    def visit_arg(self, node):
        id = get_id(node)
        if id == "self":
            return (None, "self")
        typename = "T"
        if node.annotation:
            typename = self._typename_from_annotation(node)
        return (typename, id)

    def visit_Lambda(self, node):
        _, args = self.visit(node.args)
        args_string = ", ".join(args)
        body = self.visit(node.body)
        return f"({args_string}) => {body}"

    def visit_Attribute(self, node):
        attr = node.attr

        value_id = self.visit(node.value)

        if is_list(node.value):
            if node.attr == "append":
                attr = "add"
        if not value_id:
            value_id = ""

        if is_class_or_module(value_id, node.scopes):
            return f"{value_id}.{attr}"

        if is_self_arg(value_id, node.scopes):
            return attr

        return f"{value_id}.{attr}"

    def visit_range(self, node, vargs: List[str]) -> str:
        start = 0
        step = 1
        if len(vargs) == 1:
            end = vargs[0]
        elif len(node.args) == 2:
            start = vargs[0]
            end = vargs[1]
        elif len(node.args) == 3:
            start = vargs[0]
            end = vargs[1]
            step = vargs[2]
        else:
            raise Exception(
                "encountered range() call with unknown parameters: range({})".format(
                    vargs
                )
            )

        return f"([for(var i = {start}; i < {end}; i += {step}) i])"

    def visit_print(self, node, vargs: List[str]) -> str:
        placeholders = []
        for n in node.args:
            placeholders.append("%s")
        self._usings.add("package:sprintf/sprintf.dart")
        placeholders_str = " ".join(placeholders)
        vargs_str = ", ".join(vargs)
        return rf'print(sprintf("{placeholders_str}", [{vargs_str}]))'

    def _dispatch(self, node, fname: str, vargs: List[str]) -> Optional[str]:
        dispatch_map = {
            "range": self.visit_range,
            "xrange": self.visit_range,
            "print": self.visit_print,
        }

        if fname in dispatch_map:
            return dispatch_map[fname](node, vargs)

        def visit_min_max(is_max: bool) -> str:
            min_max = "max" if is_max else "min"
            self._usings.add("dart:math")
            vargs_str = ", ".join(vargs)
            return f"{min_max}({vargs_str})"

        # small one liners are inlined here as lambdas
        small_dispatch_map = {
            "int": lambda: f"{vargs[0]}.toInt()",
            "str": lambda: f"{vargs[0]}.toString()",
            "len": lambda: f"{vargs[0]}.length",
            "floor": lambda: f"{vargs[0]}.floor()",
            "max": functools.partial(visit_min_max, is_max=True),
            "min": functools.partial(visit_min_max, is_min=True),
        }

        if fname in small_dispatch_map:
            return small_dispatch_map[fname]()
        return None

    def visit_Call(self, node):
        fname = self.visit(node.func)
        vargs = []

        if node.args:
            vargs += [self.visit(a) for a in node.args]
        if node.keywords:
            vargs += [self.visit(kw.value) for kw in node.keywords]

        ret = self._dispatch(node, fname, vargs)
        if ret is not None:
            return ret
        if vargs:
            args = ", ".join(vargs)
        else:
            args = ""
        return f"{fname}({args})"

    def visit_For(self, node):
        target = self.visit(node.target)
        it = self.visit(node.iter)
        buf = []
        buf.append("for (final {0} in {1}) {{".format(target, it))
        buf.extend([self.visit(c) for c in node.body])
        buf.append("}")
        return "\n".join(buf)

    def visit_Str(self, node):
        return "" + super().visit_Str(node) + ""

    def visit_Bytes(self, node):
        bytes_str = "{0}".format(node.s)
        return bytes_str.replace("'", '"')  # replace single quote with double quote

    def visit_Compare(self, node):
        left = self.visit(node.left)
        right = self.visit(node.comparators[0])
        if isinstance(node.ops[0], ast.In):
            return "{0}.contains({1})".format(right, left)
        elif isinstance(node.ops[0], ast.NotIn):
            return "!({0}.contains({1}))".format(right, left)

        return super().visit_Compare(node)

    def visit_Name(self, node):
        exception_name_map = {"ZeroDivisionError": "IntegerDivisionByZeroException"}
        if node.id == "None":
            return "null"
        elif node.id in exception_name_map:
            return exception_name_map[node.id]
        else:
            return super().visit_Name(node)

    def visit_NameConstant(self, node):
        if node.value is None:
            return "null"
        else:
            return super().visit_NameConstant(node)

    def _make_block(self, node):
        buf = []
        buf.append("{")
        buf.extend([self.visit(child) for child in node.body])
        buf.append("}")
        return "\n".join(buf)

    def visit_If(self, node):
        body_vars = set([get_id(v) for v in node.scopes[-1].body_vars])
        orelse_vars = set([get_id(v) for v in node.scopes[-1].orelse_vars])
        node.common_vars = body_vars.intersection(orelse_vars)

        var_definitions = []
        for cv in node.common_vars:
            typename = "var"
            if hasattr(cv, "annotation"):
                typename = get_id(cv.annotation)

            var_definitions.append((typename, cv))
        decls = "\n".join([f"{typename} {cv};" for typename, cv in var_definitions])

        return decls + "\n\n" + super().visit_If(node)

    def visit_UnaryOp(self, node):
        if isinstance(node.op, ast.USub):
            if isinstance(node.operand, (ast.Call, ast.Num)):
                # Shortcut if parenthesis are not needed
                return "-{0}".format(self.visit(node.operand))
            else:
                return "-({0})".format(self.visit(node.operand))
        else:
            return super().visit_UnaryOp(node)

    def visit_BinOp(self, node):
        if (
            isinstance(node.left, ast.List)
            and isinstance(node.op, ast.Mult)
            and isinstance(node.right, ast.Num)
        ):
            return "std::vector ({0},{1})".format(
                self.visit(node.right), self.visit(node.left.elts[0])
            )
        else:
            return super().visit_BinOp(node)

    def visit_ClassDef(self, node):
        extractor = DeclarationExtractor(DartTranspiler())
        extractor.visit(node)
        declarations = node.declarations = extractor.get_declarations()
        node.class_assignments = extractor.class_assignments

        ret = super().visit_ClassDef(node)
        if ret is not None:
            return ret

        fields = []
        index = 0
        constructor = ""
        for declaration, typename in declarations.items():
            if typename == None:
                typename = "ST{0}".format(index)
                index += 1
            fields.append(f"{typename} {declaration};")

        for b in node.body:
            if isinstance(b, ast.FunctionDef):
                b.self_type = node.name

        fields = "\n".join(fields)
        body = [self.visit(b) for b in node.body]
        if node.is_dataclass:
            args = [arg for arg in declarations.keys()]
            args = ", ".join([f"this.{arg}" for arg in args])
            constructor = f"{node.name}({args}) {{}}"
            body = [constructor] + body
        body = "\n".join(body)

        return f"class {node.name} {{\n{fields}\n\n {body}\n}}\n"

    def visit_IntEnum(self, node):
        # TODO: Consider using Vnum instead of the language built-in
        # Any python enum which specifies a non default value will break this
        fields = []
        for member, var in node.class_assignments.items():
            var = self.visit(var)
            if var == "auto()":
                fields.append(f"{member},")
            else:
                fields.append(f"{member} = {var},")
        fields = "\n".join(fields)
        return f"enum {node.name} {{\n{fields}\n}}\n\n"

    def visit_StrEnum(self, node):
        # TODO: Dedup with other enum implementations
        self._usings.add("package:vnum/vnum.dart")
        fields = []
        ename = node.name
        for i, (member, var) in enumerate(node.class_assignments.items()):
            var = self.visit(var)
            if var == "auto()":
                var = f'"{member}"'
            fields.append(
                f"            static final {member} = const {ename}.define({var});"
            )
        fields = "\n".join(["    " * 2 + f for f in fields])
        return textwrap.dedent(
            f"""\
            @VnumDefinition
            class {node.name} extends Vnum<String> {{\n{fields}

            const {node.name}.define(String fromValue) : super.define(fromValue);
              factory {node.name}(String value) => Vnum.fromValue(value, {node.name});
            }}
            """
        )

    def visit_IntFlag(self, node):
        self._usings.add("package:vnum/vnum.dart")
        fields = []
        ename = node.name
        for i, (member, var) in enumerate(node.class_assignments.items()):
            var = self.visit(var)
            if var == "auto()":
                var = i
            fields.append(
                f"            static final {member} = const {ename}.define({var});"
            )
        fields = "\n".join(["    " * 2 + f for f in fields])
        return textwrap.dedent(
            f"""\
            @VnumDefinition
            class {node.name} extends Vnum<int> {{\n{fields}

            const {node.name}.define(int fromValue) : super.define(fromValue);
              factory {node.name}(int value) => Vnum.fromValue(value, {node.name});
            }}
            """
        )

    def visit_alias(self, node):
        return "use {0}".format(node.name)

    def visit_Import(self, node):
        imports = [self.visit(n) for n in node.names]
        return "\n".join(i for i in imports if i)

    def visit_ImportFrom(self, node):
        if node.module in self.IGNORED_MODULE_LIST:
            return ""

        names = [n.name for n in node.names]
        names = ", ".join(names)
        module_path = node.module.replace(".", "::")
        return "use {0}::{{{1}}}".format(module_path, names)

    def visit_List(self, node):
        if len(node.elts) > 0:
            elements = [self.visit(e) for e in node.elts]
            return "[{0}]".format(", ".join(elements))

        else:
            return "[]"

    def visit_Dict(self, node):
        keys = [self.visit(k) for k in node.keys]
        values = [self.visit(k) for k in node.values]
        kv_pairs = ", ".join([f"{k} : {v}" for k, v in zip(keys, values)])
        return f"{{{kv_pairs}}}"

    def visit_Subscript(self, node):
        value = self.visit(node.value)
        index = self.visit(node.slice)
        if hasattr(node, "is_annotation"):
            if value in self.CONTAINER_TYPE_MAP:
                value = self.CONTAINER_TYPE_MAP[value]
            if value == "Tuple":
                return "({0})".format(index)
            return "{0}<{1}>".format(value, index)
        return "{0}[{1}]".format(value, index)

    def visit_Index(self, node):
        return self.visit(node.value)

    def visit_Slice(self, node):
        lower = ""
        if node.lower:
            lower = self.visit(node.lower)
        upper = ""
        if node.upper:
            upper = self.visit(node.upper)

        return "{0}..{1}".format(lower, upper)

    def visit_Tuple(self, node):
        elts = [self.visit(e) for e in node.elts]
        elts = ", ".join(elts)
        if hasattr(node, "is_annotation"):
            return elts
        return "({0})".format(elts)

    def visit_unsupported_body(self, name, body):
        buf = ["let {0} = {{ //unsupported".format(name)]
        buf += [self.visit(n) for n in body]
        buf.append("};")
        return buf

    def visit_Raise(self, node):
        exc = self.visit(node.exc)
        return f"throw new {exc};"

    def visit_Try(self, node):
        buf = []
        buf.append("try {")
        buf.extend([self.visit(child) for child in node.body])
        buf.append("}")

        for handler in node.handlers:
            buf.append(self.visit(handler))

        if node.finalbody:
            buf.append("finally {")
            buf.extend([self.visit(child) for child in node.finalbody])
            buf.append("}")

        return "\n".join(buf)

    def visit_ExceptHandler(self, node):
        buf = []
        if node.type:
            type_str = self.visit(node.type)
            catch_str = ""
            if node.name:
                catch_str = f" catch({node.name})"
            buf.append(f"on {type_str}{catch_str} {{")
        else:
            buf.append("catch(e) {")
        buf.extend([self.visit(child) for child in node.body])
        buf.append("}")
        return "\n".join(buf)

    def visit_Assert(self, node):
        condition = self.visit(node.test)
        return f"assert({condition});"

    def visit_AnnAssign(self, node):
        target, type_str, val = super().visit_AnnAssign(node)
        return f"{type_str} {target} = {val};"

    def _visit_AssignOne(self, node, target):
        kw = "var" if is_mutable(node.scopes, get_id(target)) else "final"

        if isinstance(target, ast.Tuple):
            self._usings.add("package:tuple/tuple.dart")
            elts = [self.visit(e) for e in target.elts]
            value = self.visit(node.value)
            value_types = "int, int"
            count = len(elts)
            tmp_var = self._get_temp()
            buf = [f"{kw} {tmp_var} = Tuple{count}<{value_types}>{value};"]
            for i, elt in enumerate(elts):
                buf.extend([f"{elt} = {tmp_var}.item{i+1};"])
            return "\n".join(buf)

        if isinstance(node.scopes[-1], ast.If):
            outer_if = node.scopes[-1]
            target_id = self.visit(target)
            if target_id in outer_if.common_vars:
                value = self.visit(node.value)
                return f"{kw} {target_id} = {value};"

        if isinstance(target, ast.Subscript) or isinstance(target, ast.Attribute):
            target = self.visit(target)
            value = self.visit(node.value)
            return f"{target} = {value};"

        definition = node.scopes.find(target.id)
        if isinstance(target, ast.Name) and defined_before(definition, node):
            target = self.visit(target)
            value = self.visit(node.value)
            return f"{target} = {value};"
        elif isinstance(node.value, ast.List):
            elements = [self.visit(e) for e in node.value.elts]
            elements = ", ".join(elements)
            target = self.visit(target)

            return f"{kw} {target} = [{elements}];"
        else:
            typename = self._typename_from_annotation(target)
            target = self.visit(target)
            value = self.visit(node.value)

            if typename != self._default_type:
                if kw == self._default_type:
                    return f"{typename} {target} = {value};"
            else:
                return f"{kw} {target} = {value};"

            return f"{kw} {typename} {target} = {value};"

    def visit_Delete(self, node):
        target = node.targets[0]
        return "{0}.drop()".format(self.visit(target))

    def visit_With(self, node):
        buf = []

        with_statement = "// with!("
        for i in node.items:
            if i.optional_vars:
                with_statement += "{0} as {1}, ".format(
                    self.visit(i.context_expr), self.visit(i.optional_vars)
                )
            else:
                with_statement += "{0}, ".format(self.visit(i.context_expr))
        with_statement = with_statement[:-2] + ") //unsupported\n{"
        buf.append(with_statement)

        for n in node.body:
            buf.append(self.visit(n))

            buf.append("}")

        return "\n".join(buf)

    def visit_Await(self, node):
        return "await!({0})".format(self.visit(node.value))

    def visit_AsyncFunctionDef(self, node):
        return "#[async]\n{0}".format(self.visit_FunctionDef(node))

    def visit_Yield(self, node):
        return "//yield is unimplemented"

    def visit_Print(self, node):
        buf = []
        for n in node.values:
            value = self.visit(n)
            buf.append('println("{{:?}}",{0})'.format(value))
        return "\n".join(buf)

    def visit_DictComp(self, node):
        return "DictComp /*unimplemented()*/"

    def visit_GeneratorExp(self, node):
        elt = self.visit(node.elt)
        generator = node.generators[0]
        target = self.visit(generator.target)
        iter = self.visit(generator.iter)

        # HACK for dictionary iterators to work
        if not iter.endswith("keys()") or iter.endswith("values()"):
            iter += ".iter()"

        map_str = ".map(|{0}| {1})".format(target, elt)
        filter_str = ""
        if generator.ifs:
            filter_str = ".cloned().filter(|&{0}| {1})".format(
                target, self.visit(generator.ifs[0])
            )

        return "{0}{1}{2}.collect::<Vec<_>>()".format(iter, filter_str, map_str)

    def visit_ListComp(self, node):
        return self.visit_GeneratorExp(node)  # right now they are the same

    def visit_Global(self, node):
        return "//global {0}".format(", ".join(node.names))

    def visit_Starred(self, node):
        return "starred!({0})/*unsupported*/".format(self.visit(node.value))

    def visit_Set(self, node):
        elements = [self.visit(e) for e in node.elts]
        elements_str = ", ".join(elements)
        return f"new Set.from([{elements_str}])"

    def visit_IfExp(self, node):
        body = self.visit(node.body)
        orelse = self.visit(node.orelse)
        test = self.visit(node.test)
        return f"{test}? ({body}) : ({orelse})"



from enum import IntEnum, IntFlag, auto


class Colors(IntEnum):
    RED = auto()
    GREEN = auto()
    BLUE = auto()


class Permissions(IntFlag):
    R = 1
    W = 2
    X = 16


def show():
    color_map = {Colors.RED: "red", Colors.GREEN: "green", Colors.BLUE: "blue"}
    a = Colors.GREEN
    if a == Colors.GREEN:
        print("green")
    else:
        print("Not green")
    b = Permissions.R
    if b == Permissions.R:
        print("R")
    else:
        print("Not R")
    assert len(color_map) != 0


if __name__ == "__main__":
    show()



import ast

from .inference import KT_TYPE_MAP, KT_WIDTH_RANK

from py2many.analysis import get_id
from py2many.clike import CLikeTranspiler as CommonCLikeTranspiler


# allowed as names in Python but treated as keywords in Kotlin
kotlin_keywords = frozenset(
    [
        "abstract",
        "actual",
        "annotation",
        "as",
        "break",
        "by",
        "catch",
        "class",
        "companion",
        "const",
        "constructor",
        "continue",
        "crossinline",
        "data",
        "delegate",
        "do",
        "dynamic",
        "else",
        "enum",
        "expect",
        "external",
        "field",
        "file",
        "file",
        "final",
        "finally",
        "for",
        "fun",
        "get",
        "if",
        "import",
        "in",
        "infix",
        "init",
        "inline",
        "inner",
        "interface",
        "internal",
        "is",
        "lateinit",
        "noinline",
        "object",
        "open",
        "operator",
        "out",
        "override",
        "package",
        "param",
        "private",
        "property",
        "protected",
        "public",
        "receiver",
        "reified",
        "return",
        "sealed",
        "set",
        "setparam",
        "super",
        "suspend",
        "tailrec",
        "this",
        "throw",
        "try",
        "typealias",
        "typeof",
        "val",
        "var",
        "vararg",
        "when",
        "where",
        "while",
    ]
)


class CLikeTranspiler(CommonCLikeTranspiler):
    def __init__(self):
        super().__init__()
        self._type_map = KT_TYPE_MAP
        self._statement_separator = ""
        self._temp = 0

    def _get_temp(self):
        self._temp += 1
        return f"__tmp{self._temp}"

    def _check_keyword(self, name):
        if name in kotlin_keywords:
            return name + "_", True
        if name == "_":
            return self._get_temp(), True
        return name, False

    def visit_Name(self, node):
        node_id, changed = self._check_keyword(node.id)
        if changed:
            return node_id
        return super().visit_Name(node)

    def visit_BinOp(self, node):
        if isinstance(node.op, ast.Pow):
            return "pow({0}, {1})".format(self.visit(node.left), self.visit(node.right))

        left = self.visit(node.left)
        op = self.visit(node.op)
        right = self.visit(node.right)

        left_type = self._typename_from_annotation(node.left)
        right_type = self._typename_from_annotation(node.right)

        left_rank = KT_WIDTH_RANK.get(left_type, -1)
        right_rank = KT_WIDTH_RANK.get(right_type, -1)

        if left_rank > right_rank:
            right = f"{right}.to{left_type}()"
        elif right_rank > left_rank:
            left = f"{left}.to{right_type}()"

        # Multiplication and division binds tighter (has higher precedence) than addition and subtraction.
        # To visually communicate this we omit spaces when multiplying and dividing.
        if isinstance(node.op, (ast.Mult, ast.Div)):
            return f"({left}{op}{right})"
        else:
            return f"({left} {op} {right})"

    def visit_Compare(self, node):
        if isinstance(node.ops[0], ast.In):
            return self.visit_In(node)

        node_right = node.comparators[0]
        left_type = self._typename_from_annotation(node.left)
        right_type = self._typename_from_annotation(node_right)

        left = self.visit(node.left)
        op = self.visit(node.ops[0])
        right = self.visit(node_right)

        left_rank = KT_WIDTH_RANK.get(left_type, -1)
        right_rank = KT_WIDTH_RANK.get(right_type, -1)

        if left_rank > right_rank and right_rank != -1:
            right = f"{right}.to{left_type}()"
        elif right_rank > left_rank and left_rank != -1:
            left = f"{left}.to{right_type}()"

        return f"{left} {op} {right}"

    def visit_In(self, node):
        left = self.visit(node.left)
        right = self.visit(node.comparators[0])
        return f"{left} in {right}"

    def _recursive_expand(self, slice_value):
        if isinstance(slice_value, ast.Name):
            slice_value = get_id(slice_value)
        elif isinstance(slice_value, (ast.Tuple, ast.List)):
            slice_value = [self._recursive_expand(e) for e in slice_value.elts]
        return slice_value

    def _typename_from_annotation(self, node, attr="annotation") -> str:
        if hasattr(node, attr):
            typename = getattr(node, attr)
            if isinstance(typename, ast.Subscript):
                if get_id(typename.value) == "Callable":
                    slice_value = self._slice_value(typename)
                    slice_value = self._recursive_expand(slice_value)

                    def recursive_tuple(lst) -> tuple:
                        return (
                            tuple(recursive_tuple(x) for x in lst)
                            if isinstance(lst, list)
                            else lst
                        )

                    slice_value = recursive_tuple(slice_value)
                    if len(slice_value) == 2:
                        # Kotlin lambda syntax
                        args = ", ".join(self._map_types(slice_value[0]))
                        ret = self._map_type(slice_value[1])
                        return f"({args}) -> {ret}"
                    return f"{slice_value}"

        return super()._typename_from_annotation(node, attr)



#!/usr/bin/env python3

import os.path
import sys

try:
    project_root = os.path.dirname(__file__)
    sys.path.insert(0, project_root)
except Exception:
    project_root = None

from py2many.cli import main


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

from setuptools import setup, find_packages

__version__ = "0.2.0"

install_requires = []
setup_requires = []
tests_require = ["pytest", "unittest-expander"]

with open("README.md") as readme_file:
    readme = readme_file.read()

setup(
    name="py2many",
    version=__version__,
    description="Python to CLike language transpiler.",
    long_description=readme + "\n\n",
    long_description_content_type = "text/markdown",
    author="Arun Sharma",
    python_requires=">=3.8",
    url="http://github.com/adsharma/py2many",
    install_requires=install_requires,
    setup_requires=setup_requires,
    tests_require=tests_require,
    packages=find_packages(exclude=["docs", "examples", "tests", "tests*"]),
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Topic :: Software Development",
        "Topic :: Utilities",
    ],
    test_suite="tests",
    entry_points={
        "console_scripts": ["py2many=py2many.cli:main"],
    },
)



#!/usr/bin/env python3

from ctypes import (
    c_int8,
    c_int16,
    c_int32,
    c_int64,
    c_uint8,
    c_uint16,
    c_uint32,
    c_uint64,
)


def foo():
    a = 10
    b = 20
    _c1 = a + b
    _c2 = a - b
    _c3 = a * b
    _c4 = a / b
    _c5 = -a
    d = 2.0
    _e1 = a + d
    _e2 = a - d
    _e3 = a * d
    _e4 = a / d
    _f = -3.0
    _g = -a


def add1(x: c_int8, y: c_int8):
    return x + y


def add2(x: c_int16, y: c_int16):
    return x + y


def add3(x: c_int32, y: c_int32):
    return x + y


def add4(x: c_int64, y: c_int64):
    return x + y


def add5(x: c_uint8, y: c_uint8):
    return x + y


def add6(x: c_uint16, y: c_uint16):
    return x + y


def add7(x: c_uint32, y: c_uint32):
    return x + y


def add8(x: c_uint64, y: c_uint64):
    return x + y


def add9(x: c_int8, y: c_uint16):
    return x + y


def sub(x: c_int8, y: c_int8):
    return x - y


def mul(x: c_int8, y: c_int8):
    return x * y


def fadd1(x: c_int8, y: float):
    return x + y


def show():
    rv = fadd1(6, 6.0)
    assert rv == 12
    print("OK")


if __name__ == "__main__":
    foo()
    show()



import sys
import ast


def add_imports(node):
    """Provide context of imports Module"""
    return ImportTransformer().visit(node)


def is_void_function(fun):
    finder = ReturnFinder()
    finder.visit(fun)
    return not (finder.returns or fun.returns is not None)


if sys.version_info[0] >= 3:

    def get_id(var):
        if isinstance(var, ast.alias):
            return var.name
        elif isinstance(var, ast.Name):
            return var.id
        elif isinstance(var, ast.arg):
            return var.arg
        elif isinstance(var, ast.FunctionDef):
            return var.name
        elif isinstance(var, ast.ClassDef):
            return var.name
        else:
            # print(f"warning: {var}"")
            return None


else:

    def get_id(var):
        if isinstance(var, ast.alias):
            return var.name
        elif isinstance(var, ast.Name):
            return var.id


def is_global(target):
    return isinstance(target.scopes[-1], ast.Module)


def is_mutable(scopes, target):
    for scope in scopes:
        if isinstance(scope, ast.FunctionDef):
            if target in scope.mutable_vars:
                return True
    return False


class ReturnFinder(ast.NodeVisitor):
    returns = False

    def visit_Return(self, node):
        if node.value != None:
            self.returns = True


class FunctionTransformer(ast.NodeTransformer):
    """Tracks defined functions in scope"""

    def visit_Module(self, node):
        node.defined_functions = []
        self.generic_visit(node)
        return node

    def visit_FunctionDef(self, node):
        node.defined_functions = []
        node.scopes[-2].defined_functions.append(node)
        self.generic_visit(node)
        return node

    def visit_ImportFrom(self, node):
        for name in node.names:
            node.scopes[-1].defined_functions.append(name)
        return node


class CalledWithTransformer(ast.NodeTransformer):
    """
    Tracks whether variables or functions get
    used as arguments of other functions
    """

    def visit_Assign(self, node):
        for target in node.targets:
            target.called_with = []
        return node

    def visit_FunctionDef(self, node):
        node.called_with = []
        self.generic_visit(node)
        return node

    def visit_Call(self, node):
        for arg in node.args:
            if isinstance(arg, ast.Name):
                var = node.scopes.find(arg.id)
                var.called_with.append(node)
        self.generic_visit(node)
        return node


class AttributeCallTransformer(ast.NodeTransformer):
    """Tracks attribute function calls on variables"""

    def visit_Assign(self, node):
        for target in node.targets:
            target.calls = []
        return node

    def visit_Call(self, node):
        if isinstance(node.func, ast.Attribute):
            var = node.scopes.find(node.func.value.id)
            var.calls.append(node)
        return node


class ImportTransformer(ast.NodeTransformer):
    """Adds imports to scope block"""

    def visit_ImportFrom(self, node):
        for name in node.names:
            name.imported_from = node
            scope = name.scopes[-1]
            if hasattr(scope, "imports"):
                scope.imports.append(name)
        return node

    def visit_Module(self, node):
        node.imports = []
        self.generic_visit(node)
        return node



import ast
from contextlib import contextmanager

from py2many.analysis import get_id


def add_scope_context(node):
    """Provide to scope context to all nodes"""
    return ScopeTransformer().visit(node)


class ScopeMixin(object):
    """
    Adds a scope property with the current scope (function, module)
    a node is part of.
    """

    scopes = []

    @contextmanager
    def enter_scope(self, node):
        if self._is_scopable_node(node):
            self.scopes.append(node)
            yield
            self.scopes.pop()
        else:
            yield

    @property
    def scope(self):
        try:
            return self.scopes[-1]
        except IndexError:
            return None

    def _is_scopable_node(self, node):
        scopes = [ast.Module, ast.ClassDef, ast.FunctionDef, ast.For, ast.If, ast.With]
        return len([s for s in scopes if isinstance(node, s)]) > 0


class ScopeList(list):
    """
    Wraps around list of scopes and provides find method for finding
    the definition of a variable
    """

    def find(self, lookup):
        """Find definition of variable lookup."""

        def find_definition(scope, var_attr="vars"):
            for var in getattr(scope, var_attr):
                if get_id(var) == lookup:
                    return var

        for scope in self:
            defn = None
            if not defn and hasattr(scope, "vars"):
                defn = find_definition(scope, "vars")
            if not defn and hasattr(scope, "body_vars"):
                defn = find_definition(scope, "body_vars")
            if not defn and hasattr(scope, "orelse_vars"):
                defn = find_definition(scope, "orelse_vars")
            if not defn and hasattr(scope, "body"):
                defn = find_definition(scope, "body")
            if defn:
                return defn

    def find_import(self, lookup):
        for scope in reversed(self):
            if hasattr(scope, "imports"):
                for imp in scope.imports:
                    if imp.name == lookup:
                        return imp


class ScopeTransformer(ast.NodeTransformer, ScopeMixin):
    """
    Adds a scope attribute to each node.
    The scope contains the current scope (function, module, for loop)
    a node is part of.
    """

    def visit(self, node):
        with self.enter_scope(node):
            node.scopes = ScopeList(self.scopes)
            return super().visit(node)



import ast
import functools
import textwrap

from typing import Optional, List

from .clike import CLikeTranspiler
from .declaration_extractor import DeclarationExtractor
from .inference import get_inferred_go_type

from py2many.analysis import get_id, is_global, is_void_function
from py2many.rewriters import capitalize_first, rename, camel_case
from py2many.tracer import is_list, defined_before, is_class_or_module, is_enum


class GoMethodCallRewriter(ast.NodeTransformer):
    def visit_Call(self, node):
        needs_assign = False
        fname = node.func
        if isinstance(fname, ast.Attribute):
            if is_list(node.func.value) and fname.attr == "append":
                needs_assign = True
            if get_id(fname.value):
                node0 = ast.Name(id=get_id(fname.value), lineno=node.lineno)
            else:
                node0 = fname.value
            node.args = [node0] + node.args
            node.func = ast.Name(id=fname.attr, lineno=node.lineno, ctx=fname.ctx)
        if needs_assign:
            ret = ast.Assign(
                targets=[ast.Name(id=fname.value.id, lineno=node.lineno)],
                value=node,
                lineno=node.lineno,
                scopes=node.scopes,
            )
            return ret
        return node


class GoNoneCompareRewriter(ast.NodeTransformer):
    def visit_Compare(self, node):
        left = self.visit(node.left)
        right = self.visit(node.comparators[0])
        if (
            isinstance(right, ast.Constant)
            and right.value is None
            and isinstance(left, ast.Constant)
            and isinstance(left.value, int)
        ):
            # Convert None to 0 to compare vs int
            right.value = 0
        return node


class GoPropagateTypeAnnotation(ast.NodeTransformer):
    def _visit_assign(self, node, target):
        if hasattr(node, "annotation") and isinstance(
            node.value, (ast.List, ast.Set, ast.Dict)
        ):
            node.value.annotation = node.annotation
        return node

    def visit_Assign(self, node):
        target = node.targets[0]
        return self._visit_assign(node, target)

    def visit_AnnAssign(self, node):
        target = node.target
        return self._visit_assign(node, target)


class GoVisibilityRewriter(ast.NodeTransformer):
    def visit_Name(self, node):
        if hasattr(node, "scopes") and is_global(node):
            old_name = get_id(node)
            new_name = camel_case(old_name)
            if old_name != new_name:
                rename(node.scopes[-1], old_name, new_name)
        return node

    def visit_FunctionDef(self, node):
        if hasattr(node, "scopes") and isinstance(node.scopes[-2], ast.Module):
            old_name = get_id(node)
            if old_name is not None:
                if "_" in old_name:
                    new_name = camel_case(old_name)
                else:
                    new_name = capitalize_first(old_name)
                if old_name != new_name:
                    rename(node.scopes[-2], old_name, new_name)
        return node


class GoIfExpRewriter(ast.NodeTransformer):
    def visit_Assign(self, node):
        if isinstance(node.value, ast.IfExp):
            if_stmt = ast.parse(
                textwrap.dedent(
                    """\
                if True:
                    a = 1
                else:
                    a = 2"""
                )
            ).body[0]
            assert isinstance(if_stmt, ast.If)
            if_stmt.test = node.value.test
            if_stmt.body[0].targets = node.targets
            if_stmt.body[0].value = node.value.body
            if_stmt.orelse[0].targets = node.targets
            if_stmt.orelse[0].value = node.value.orelse
            if_stmt.lineno = node.lineno
            if_stmt.col_offset = node.col_offset
            ast.fix_missing_locations(if_stmt)
            return if_stmt
        return node


class GoTranspiler(CLikeTranspiler):
    NAME = "go"

    CONTAINER_TYPE_MAP = {"List": "[]", "Dict": None, "Set": None, "Optional": "nil"}

    def __init__(self):
        super().__init__()
        self._default_type = None
        self._container_type_map = self.CONTAINER_TYPE_MAP

    def headers(self, meta):
        return "\n".join(self._headers)

    def usings(self):
        buf = "package main\n\n"  # TODO naming
        if self._usings:
            buf += "import (\n"
            buf += "\n".join([f"{using}" for using in self._usings])
            buf += ")\n"
        return buf + "\n\n"

    def _combine_value_index(self, value_type, index_type) -> str:
        return f"{value_type}{index_type}"

    def comment(self, text):
        return f"// {text}\n"

    def visit_FunctionDef(self, node):
        body = "\n".join([self.visit(n) for n in node.body])
        typenames, args = self.visit(node.args)

        if len(typenames) and typenames[0] == None and hasattr(node, "self_type"):
            typenames[0] = node.self_type

        args_list = []
        typedecls = []
        index = 0
        for i in range(len(args)):
            typename = typenames[i]
            arg = args[i]
            if typename == "T":
                typename = "T{0} any".format(index)
                typedecls.append(typename)
                index += 1
            args_list.append(f"{arg} {typename}")

        return_type = ""
        if not is_void_function(node):
            if node.returns:
                typename = self._typename_from_annotation(node, attr="returns")
                return_type = f" {typename}"
            else:
                return_type = " RT"
                typedecls.append("RT")

        template = ""
        if len(typedecls) > 0:
            template = "[{0}]".format(", ".join(typedecls))

        args = ", ".join(args_list)
        funcdef = f"func {node.name}{template}({args}){return_type} {{"
        return f"{funcdef}\n{body}}}\n\n"

    def visit_Return(self, node):
        if node.value:
            ret = self.visit(node.value)
            fndef = None
            for scope in node.scopes:
                if isinstance(scope, ast.FunctionDef):
                    fndef = scope
                    break
            if fndef:
                return_type = self._typename_from_annotation(fndef, attr="returns")
                value_type = get_inferred_go_type(node.value)
                if return_type != value_type and value_type is not None:
                    return f"return {return_type}({ret})"
            return f"return {ret}"
        return "return"

    def visit_arg(self, node):
        id = get_id(node)
        if id == "self":
            return (None, "self")
        typename = "T"
        if node.annotation:
            typename = self._typename_from_annotation(node)
        return (typename, id)

    def visit_Lambda(self, node):
        typenames, args = self.visit(node.args)
        # HACK: to pass unit tests. TODO: infer types
        # Need to get it from the annotation on the lhs of `node`
        typenames = ["int"] * len(args)
        return_type = "int"
        args = [f"{name} {typename}" for name, typename in zip(args, typenames)]
        args_string = ", ".join(args)
        body = self.visit(node.body)
        return f"func({args_string}) {return_type} {{ return {body} }}"

    def visit_Attribute(self, node):
        attr = node.attr

        value_id = self.visit(node.value)

        if not value_id:
            value_id = ""

        if is_enum(value_id, node.scopes):
            return f"{attr}"

        if is_class_or_module(value_id, node.scopes):
            return f"{value_id}.{attr}"

        return f"{value_id}.{attr}"

    def visit_range(self, node, vargs: List[str]) -> str:
        self._usings.add('iter "github.com/hgfischer/go-iter"')
        if len(node.args) == 1:
            return f"iter.NewIntSeq(iter.Start(0), iter.Stop({vargs[0]})).All()"
        elif len(node.args) == 2:
            return (
                f"iter.NewIntSeq(iter.Start({vargs[0]}), iter.Stop({vargs[1]})).All()"
            )
        elif len(node.args) == 3:
            return f"iter.NewIntSeq(iter.Start({vargs[0]}), iter.Stop({vargs[1]}), iter.Step({vargs[2]})).All()"

        raise Exception(
            "encountered range() call with unknown parameters: range({})".format(vargs)
        )

    def visit_print(self, node, vargs: List[str]) -> str:
        placeholders = []
        for n in node.args:
            placeholders.append("%v")
        self._usings.add('"fmt"')
        placeholders_str = " ".join(placeholders)
        vargs_str = ", ".join(vargs)
        return f'fmt.Printf("{placeholders_str}\\n",{vargs_str})'

    def _dispatch(self, node, fname: str, vargs: List[str]) -> Optional[str]:
        dispatch_map = {
            "range_": self.visit_range,
            "xrange": self.visit_range,
            "print": self.visit_print,
        }

        if fname in dispatch_map:
            return dispatch_map[fname](node, vargs)

        def visit_min_max(is_max: bool) -> str:
            min_max = "math.Max" if is_max else "math.Min"
            self._usings.add('"math"')
            vargs_str = ", ".join(vargs)
            return f"{min_max}({vargs_str})"

        def visit_floor():
            self._usings.add('"math"')
            return f"math.Floor({vargs[0]})"

        # small one liners are inlined here as lambdas
        small_dispatch_map = {
            "str": lambda: f"String({vargs[0]})",
            "max": functools.partial(visit_min_max, is_max=True),
            "min": functools.partial(visit_min_max, is_min=True),
            "floor": visit_floor,
        }
        if fname in small_dispatch_map:
            return small_dispatch_map[fname]()
        return None

    def _visit_struct_literal(self, node, fname: str, fndef: ast.ClassDef):
        vargs = []  # visited args
        if not hasattr(fndef, "declarations"):
            raise Exception("Missing declarations")
        if node.args:
            for arg, decl in zip(node.args, fndef.declaration.keys()):
                arg = self.visit(arg)
                vargs += [f"{decl}: {arg}"]
        if node.keywords:
            for kw in node.keywords:
                value = self.visit(kw.value)
                vargs += [f"{kw.arg}: {value}"]
        args = ", ".join(vargs)
        return f"{fname}{{{args}}}"

    def visit_Call(self, node):
        fname = self.visit(node.func)
        fndef = node.scopes.find(fname)

        if isinstance(fndef, ast.ClassDef):
            return self._visit_struct_literal(node, fname, fndef)

        vargs = []

        if node.args:
            vargs += [self.visit(a) for a in node.args]
        if node.keywords:
            vargs += [self.visit(kw.value) for kw in node.keywords]

        ret = self._dispatch(node, fname, vargs)
        if ret is not None:
            return ret
        if vargs:
            args = ", ".join(vargs)
        else:
            args = ""
        return f"{fname}({args})"

    def visit_For(self, node):
        target = self.visit(node.target)
        it = self.visit(node.iter)
        buf = []
        if target == "_":
            buf.append(f"for range {it} {{")
        else:
            buf.append(f"for _, {target} := range {it} {{")
            # Dummy assign to silence the compiler on unused vars
            if target.startswith("_"):
                buf.append(f"_ = {target}")
        buf.extend([self.visit(c) for c in node.body])
        buf.append("}")
        return "\n".join(buf)

    def visit_While(self, node):
        buf = []
        buf.append("for {0} {{".format(self.visit(node.test)))
        buf.extend([self.visit(c) for c in node.body])
        buf.append("}")
        return "\n".join(buf)

    def visit_Str(self, node):
        return "" + super().visit_Str(node) + ""

    def visit_Bytes(self, node):
        bytes_str = "{0}".format(node.s)
        return bytes_str.replace("'", '"')  # replace single quote with double quote

    def _visit_container_compare(self, node):
        left = self.visit(node.left)
        op = self.visit(node.ops[0])
        right = self.visit(node.comparators[0])
        if op == "==":
            return f"cmp.Equal({left}, {right})"
        return super().visit_Compare(node)

    def visit_Compare(self, node):
        left = node.left
        right = node.comparators[0]
        if hasattr(left, "annotation") or hasattr(right, "annotation"):
            self._typename_from_annotation(left)
            self._typename_from_annotation(right)
            op = self.visit(node.ops[0])
            if (
                hasattr(left, "container_type")
                or hasattr(right, "container_type")
                and op != "in"
            ):
                self._usings.add('"github.com/google/go-cmp/cmp"')
                return self._visit_container_compare(node)
        left = self.visit(node.left)
        right = self.visit(node.comparators[0])
        return super().visit_Compare(node)

    def visit_Name(self, node):
        if node.id == "None":
            return "None"
        else:
            return super().visit_Name(node)

    def visit_NameConstant(self, node):
        if node.value is None:
            return "nil"
        else:
            return super().visit_NameConstant(node)

    def _make_block(self, node):
        buf = []
        buf.append("{")
        buf.extend([self.visit(child) for child in node.body])
        buf.append("}")
        return "\n".join(buf)

    def visit_If(self, node):
        body_vars = {get_id(v): v for v in node.scopes[-1].body_vars}
        orelse_vars = {get_id(v): v for v in node.scopes[-1].orelse_vars}
        node.common_vars = set(body_vars.keys()).intersection(set(orelse_vars.keys()))
        types = [self._typename_from_annotation(body_vars[v]) for v in node.common_vars]
        common_vars = "\n".join(
            [f"var {v} {t}" for v, t in zip(node.common_vars, types)]
        )
        if common_vars:
            return common_vars + "\n" + super().visit_If(node)
        else:
            return super().visit_If(node)

    def visit_UnaryOp(self, node):
        if isinstance(node.op, ast.USub):
            if isinstance(node.operand, (ast.Call, ast.Num)):
                # Shortcut if parenthesis are not needed
                return "-{0}".format(self.visit(node.operand))
            else:
                return "-({0})".format(self.visit(node.operand))
        else:
            return super().visit_UnaryOp(node)

    def visit_BinOp(self, node):
        if (
            isinstance(node.left, ast.List)
            and isinstance(node.op, ast.Mult)
            and isinstance(node.right, ast.Num)
        ):
            return "std::vector ({0},{1})".format(
                self.visit(node.right), self.visit(node.left.elts[0])
            )
        else:
            return super().visit_BinOp(node)

    def visit_ClassDef(self, node):
        extractor = DeclarationExtractor(GoTranspiler())
        extractor.visit(node)
        declarations = node.declarations = extractor.get_declarations()
        node.class_assignments = extractor.class_assignments
        ret = super().visit_ClassDef(node)
        if ret is not None:
            return ret

        fields = []
        index = 0
        for declaration, typename in declarations.items():
            if typename == None:
                typename = "ST{0}".format(index)
                index += 1
            fields.append(f"{declaration} {typename}")

        for b in node.body:
            if isinstance(b, ast.FunctionDef):
                b.self_type = node.name

        if node.is_dataclass:
            fields = "\n".join(fields)
            body = [self.visit(b) for b in node.body]
            body = "\n".join(body)
            return f"type {node.name} struct {{\n{fields}\n}}\n{body}\n"
        else:
            fields = "\n".join(fields)
            body = [self.visit(b) for b in node.body]
            body = "\n".join(body)
            return f"class {node.name} {{\n{fields}\n\n {body}\n}}\n"

    def _visit_enum(self, node, typename, members):
        ret = f"type {node.name} {typename}\n\n"
        fields = []
        for i, (member, var) in enumerate(members):
            typename = f" {node.name}" if i == 0 else ""
            if var == "auto()":
                fields.append(f"{member}{typename} = iota")
            else:
                fields.append(f"{member}{typename} = {var}")
        fields = "\n".join(fields)
        return f"{ret} const (\n{fields}\n)\n\n"

    def visit_IntEnum(self, node):
        members = []
        for i, (member, var) in enumerate(node.class_assignments.items()):
            var = self.visit(var)
            if var == "auto()":
                members.append((member, "iota"))
            else:
                members.append((member, var))
        return self._visit_enum(node, "int", members)

    def visit_IntFlag(self, node):
        members = []
        for i, (member, var) in enumerate(node.class_assignments.items()):
            var = self.visit(var)
            if var == "auto()":
                members.append((member, "1 << iota"))
            else:
                members.append((member, var))
        return self._visit_enum(node, "int", members)

    def visit_StrEnum(self, node):
        members = []
        for i, (member, var) in enumerate(node.class_assignments.items()):
            var = self.visit(var)
            members.append((member, var))
        return self._visit_enum(node, "string", members)

    def visit_alias(self, node):
        return "use {0}".format(node.name)

    def visit_Import(self, node):
        imports = [self.visit(n) for n in node.names]
        return "\n".join(i for i in imports if i)

    def visit_ImportFrom(self, node):
        if node.module in self.IGNORED_MODULE_LIST:
            return ""

        names = [n.name for n in node.names]
        names = ", ".join(names)
        module_path = node.module.replace(".", "::")
        return "use {0}::{{{1}}}".format(module_path, names)

    def visit_List(self, node):
        _ = self._typename_from_annotation(node)
        element_type = self._default_type
        if hasattr(node, "container_type"):
            _, element_type = node.container_type
        elements = [self.visit(e) for e in node.elts]
        elements_str = ", ".join(elements)
        return f"[]{element_type}{{{elements_str}}}"

    def visit_Set(self, node):
        _ = self._typename_from_annotation(node)
        element_type = self._default_type
        if hasattr(node, "container_type"):
            _, element_type = node.container_type
        elements = [self.visit(e) for e in node.elts]
        kv_pairs = ", ".join([f"{k}: true" for k in elements])
        return f"map[{element_type}]bool{{{kv_pairs}}}"

    def visit_Dict(self, node):
        keys = [self.visit(k) for k in node.keys]
        values = [self.visit(k) for k in node.values]
        kv_pairs = ", ".join([f"{k}: {v}" for k, v in zip(keys, values)])
        _ = self._typename_from_annotation(node)
        key_typename = value_typename = self._default_type
        if hasattr(node, "container_type"):
            container_type, element_type = node.container_type
            key_typename, value_typename = element_type
            if key_typename == self._default_type:
                key_typename = "int"
        return f"map[{key_typename}]{value_typename}{{{kv_pairs}}}"

    def visit_Subscript(self, node):
        value = self.visit(node.value)
        index = self.visit(node.slice)
        if hasattr(node, "is_annotation"):
            if value in self.CONTAINER_TYPE_MAP:
                value = self.CONTAINER_TYPE_MAP[value]
            if value == "Tuple":
                return "({0})".format(index)
            return "{0}{1}".format(value, index)
        return "{0}[{1}]".format(value, index)

    def visit_Index(self, node):
        return self.visit(node.value)

    def visit_Slice(self, node):
        lower = ""
        if node.lower:
            lower = self.visit(node.lower)
        upper = ""
        if node.upper:
            upper = self.visit(node.upper)

        return "{0}..{1}".format(lower, upper)

    def visit_Tuple(self, node):
        elts = [self.visit(e) for e in node.elts]
        elts = ", ".join(elts)
        if hasattr(node, "is_annotation"):
            return elts
        return "{0}".format(elts)

    def visit_unsupported_body(self, name, body):
        buf = ["let {0} = {{ //unsupported".format(name)]
        buf += [self.visit(n) for n in body]
        buf.append("};")
        return buf

    def visit_Try(self, node, finallybody=None):
        buf = self.visit_unsupported_body("try_dummy", node.body)

        for handler in node.handlers:
            buf += self.visit(handler)
        # buf.append("\n".join(excepts));

        if finallybody:
            buf += self.visit_unsupported_body("finally_dummy", finallybody)

        return "\n".join(buf)

    def visit_ExceptHandler(self, node):
        exception_type = ""
        if node.type:
            exception_type = self.visit(node.type)
        name = "except!({0})".format(exception_type)
        body = self.visit_unsupported_body(name, node.body)
        return body

    def visit_Assert(self, node):
        condition = self.visit(node.test)
        return f'if !({condition}) {{ panic("assert") }}'

    def visit_AnnAssign(self, node):
        target = self.visit(node.target)
        type_str = self._typename_from_annotation(node)
        val = self.visit(node.value) if node.value is not None else None
        if type_str is not self._default_type:
            return f"var {target} {type_str} = {val}"
        else:
            return f"var {target} = {val}"

    def _needs_cast(self, left, right) -> bool:
        if not hasattr(left, "annotation"):
            return False
        left_type = self._typename_from_annotation(left)
        right_type = get_inferred_go_type(right)
        return left_type != right_type and (
            left_type != self._default_type and right_type != self._default_type
        )

    def _assign_cast(
        self, value_str: str, cast_to: str, python_annotation, rust_annotation
    ) -> str:
        # python/rust annotations provided to customize the cast if necessary
        return f"{cast_to}({value_str})"

    def _visit_AssignOne(self, node, target):
        if isinstance(target, ast.Tuple):
            elts = [self.visit(e) for e in target.elts]
            value = self.visit(node.value)
            return "var {0} = {1}".format(", ".join(elts), value)

        if isinstance(node.scopes[-1], ast.If):
            outer_if = node.scopes[-1]
            target_id = self.visit(target)
            if target_id in outer_if.common_vars:
                value = self.visit(node.value)
                return "{0} = {1}".format(target_id, value)

        if isinstance(target, ast.Subscript) or isinstance(target, ast.Attribute):
            target = self.visit(target)
            value = self.visit(node.value)
            if value == None:
                value = "None"
            return "{0} = {1}".format(target, value)

        typename = self._typename_from_annotation(target)
        needs_cast = self._needs_cast(target, node.value)
        target_str = self.visit(target)
        value = self.visit(node.value)
        if needs_cast:
            left_annotation = target.annotation
            right_annotation = getattr(node.value, "annotation", None)
            if right_annotation is None:
                right_annotation = ast.Name(id=get_inferred_go_type(node.value))
            value = self._assign_cast(
                value, typename, left_annotation, right_annotation
            )

        definition = node.scopes.find(target.id)
        if isinstance(target, ast.Name) and defined_before(definition, node):
            return f"{target_str} = {value}"
        else:
            if typename is not None:
                # Dummy assign to silence the compiler on unused vars
                maybe_tail = ""
                if target_str.startswith("_"):
                    maybe_tail = f"\n_ = {target_str}"
                return f"var {target_str} {typename} = {value}{maybe_tail}"

            if is_global(node):
                return f"var {target_str} = {value}"
            return f"{target_str} := {value}"

    def visit_Delete(self, node):
        target = node.targets[0]
        return "{0}.drop()".format(self.visit(target))

    def visit_Raise(self, node):
        if node.exc is not None:
            return "raise!({0}); //unsupported".format(self.visit(node.exc))
        # This handles the case where `raise` is used without
        # specifying the exception.
        return "raise!(); //unsupported"

    def visit_With(self, node):
        buf = []

        with_statement = "// with!("
        for i in node.items:
            if i.optional_vars:
                with_statement += "{0} as {1}, ".format(
                    self.visit(i.context_expr), self.visit(i.optional_vars)
                )
            else:
                with_statement += "{0}, ".format(self.visit(i.context_expr))
        with_statement = with_statement[:-2] + ") //unsupported\n{"
        buf.append(with_statement)

        for n in node.body:
            buf.append(self.visit(n))

            buf.append("}")

        return "\n".join(buf)

    def visit_Await(self, node):
        return "await!({0})".format(self.visit(node.value))

    def visit_AsyncFunctionDef(self, node):
        return "#[async]\n{0}".format(self.visit_FunctionDef(node))

    def visit_Yield(self, node):
        return "//yield is unimplemented"

    def visit_Print(self, node):
        buf = []
        self._usings.add('"fmt"')
        for n in node.values:
            value = self.visit(n)
            buf.append('fmt.Printf("%v\n",{0})'.format(value))
        return "\n".join(buf)

    def visit_DictComp(self, node):
        return "DictComp /*unimplemented()*/"

    def visit_GeneratorExp(self, node):
        elt = self.visit(node.elt)
        generator = node.generators[0]
        target = self.visit(generator.target)
        iter = self.visit(generator.iter)

        # HACK for dictionary iterators to work
        if not iter.endswith("keys()") or iter.endswith("values()"):
            iter += ".iter()"

        map_str = ".map(|{0}| {1})".format(target, elt)
        filter_str = ""
        if generator.ifs:
            filter_str = ".cloned().filter(|&{0}| {1})".format(
                target, self.visit(generator.ifs[0])
            )

        return "{0}{1}{2}.collect::<Vec<_>>()".format(iter, filter_str, map_str)

    def visit_ListComp(self, node):
        return self.visit_GeneratorExp(node)  # right now they are the same

    def visit_Global(self, node):
        return "//global {0}".format(", ".join(node.names))

    def visit_Starred(self, node):
        return "starred!({0})/*unsupported*/".format(self.visit(node.value))



from typing import List


def do_pass():
    pass


def inline_pass():
    pass


def inline_ellipsis():
    ...


def indexing():
    sum = 0
    a: List[int] = []
    for i in range(10):
        a.append(i)
        sum += a[i]
    return sum


def infer_bool(code: int):
    return code in [1, 2, 4]


def show():
    # assign
    a1 = 10
    # multi-assign
    b1 = b2 = 15
    assert b1 == 15
    assert b2 == 15
    b9 = 2
    b10 = 2
    assert b9 == b10
    # annotated assign
    a2: float = 2.1
    print(a2)
    # for loop
    for i in range(0, 10):
        print(i)
    for i in range(0, 10, 2):
        print(i)
    # for i in range(1, 10):
    #     print(i)
    # unary op
    a3 = -a1
    # binary op
    a4 = a3 + a1
    print(a4)
    # ternary op
    t1 = 10 if a1 > 5 else 5
    assert t1 == 10
    sum1 = indexing()
    print(sum1)
    # lists, sets and dict
    a5 = [1, 2, 3]
    print(len(a5))
    a9: List[str] = ["a", "b", "c", "d"]
    print(len(a9))
    a6 = {1, 2, 3, 4}
    print(len(a6))
    a7 = {"a": 1, "b": 2}
    print(len(a7))
    a8 = True
    # print()
    if a8:
        print("true")
    elif a4 > 0:
        print("never get here")
    if a1 == 11:
        print("false")
    else:
        print("true")
    if 1 != None:
        print("World is sane")
    # print(True)
    # if True: a1 += 1
    # if True: print("true")
    do_pass()
    inline_pass()
    s = "1\
    2"
    print(s)
    # assert 1 != 2 != 3


if __name__ == "__main__":
    show()



import ast

from py2many.inference import get_inferred_type, is_reference, InferTypesTransformer
from py2many.analysis import get_id, is_mutable

RUST_TYPE_MAP = {
    "int": "i32",
    "float": "f64",
    "bytes": "&[u8]",
    "str": "&str",
    "bool": "bool",
    "c_int8": "i8",
    "c_int16": "i16",
    "c_int32": "i32",
    "c_int64": "i64",
    "c_uint8": "u8",
    "c_uint16": "u16",
    "c_uint32": "u32",
    "c_uint64": "u64",
}


RUST_WIDTH_RANK = {
    "bool": 0,
    "i8": 1,
    "u8": 2,
    "i16": 3,
    "u16": 4,
    "i32": 5,
    "u32": 6,
    "i64": 7,
    "u64": 8,
    "f32": 9,
    "f64": 10,
}


def infer_rust_types(node):
    visitor = InferRustTypesTransformer()
    visitor.visit(node)


def map_type(typename):
    if typename in RUST_TYPE_MAP:
        return RUST_TYPE_MAP[typename]
    return typename


def is_rust_reference(node):
    if not is_reference(node):
        return False
    if isinstance(node, ast.Call):
        definition = node.scopes.find(get_id(node.func))
        needs_reference = getattr(definition, "rust_return_needs_reference", True)
        return needs_reference
    return True


def get_inferred_rust_type(node):
    if hasattr(node, "rust_annotation"):
        return node.rust_annotation
    if isinstance(node, ast.Name):
        if not hasattr(node, "scopes"):
            return None
        definition = node.scopes.find(get_id(node))
        # Prevent infinite recursion
        if definition != node:
            return get_inferred_rust_type(definition)
    python_type = get_inferred_type(node)
    ret = map_type(get_id(python_type))
    node.rust_annotation = ret
    return ret


class InferRustTypesTransformer(ast.NodeTransformer):
    """Implements rust type inference logic as opposed to python type inference logic"""

    FIXED_WIDTH_INTS = InferTypesTransformer.FIXED_WIDTH_INTS
    FIXED_WIDTH_INTS_NAME_LIST = InferTypesTransformer.FIXED_WIDTH_INTS_NAME
    FIXED_WIDTH_INTS_NAME = InferTypesTransformer.FIXED_WIDTH_INTS_NAME_LIST

    def __init__(self):
        super().__init__()

    def _handle_overflow(self, op, left_id, right_id):
        left_rust_id = map_type(left_id)
        right_rust_id = map_type(right_id)

        left_rust_rank = RUST_WIDTH_RANK.get(left_rust_id, -1)
        right_rust_rank = RUST_WIDTH_RANK.get(right_rust_id, -1)

        return left_id if left_rust_rank > right_rust_rank else right_id

    def visit_BinOp(self, node):
        self.generic_visit(node)

        if isinstance(node.left, ast.Name):
            lvar = node.scopes.find(get_id(node.left))
        else:
            lvar = node.left

        if isinstance(node.right, ast.Name):
            rvar = node.scopes.find(get_id(node.right))
        else:
            rvar = node.right

        left = lvar.annotation if lvar and hasattr(lvar, "annotation") else None
        right = rvar.annotation if rvar and hasattr(rvar, "annotation") else None

        if left is None and right is not None:
            node.rust_annotation = map_type(get_id(right))
            return node

        if right is None and left is not None:
            node.rust_annotation = map_type(get_id(left))
            return node

        if right is None and left is None:
            return node

        # Both operands are annotated. Now we have interesting cases
        left_id = get_id(left)
        right_id = get_id(right)

        # Special case int op int = int, where op != Div
        if (
            left_id == right_id
            and left_id == "int"
            and not isinstance(node.op, ast.Div)
        ):
            node.annotation = left
            node.go_annotation = map_type(left_id)
            return node

        if left_id == "int":
            left_id = "c_int32"
        if right_id == "int":
            right_id = "c_int32"

        if (
            left_id in self.FIXED_WIDTH_INTS_NAME
            and right_id in self.FIXED_WIDTH_INTS_NAME
        ):
            ret = self._handle_overflow(node.op, left_id, right_id)
            node.rust_annotation = map_type(ret)
            return node
        if left_id == right_id:
            node.annotation = left
            node.rust_annotation = map_type(left_id)
            return node
        else:
            if left_id in self.FIXED_WIDTH_INTS_NAME:
                left_id = "int"
            if right_id in self.FIXED_WIDTH_INTS_NAME:
                right_id = "int"
            if (left_id, right_id) in {("int", "float"), ("float", "int")}:
                node.rust_annotation = map_type("float")
                return node

            raise Exception(f"type error: {left_id} {type(node.op)} {right_id}")

    def visit_Return(self, node):
        self.generic_visit(node)
        if node.value:
            fndef = None
            for scope in node.scopes:
                if isinstance(scope, ast.FunctionDef):
                    fndef = scope
                    break
            if fndef:
                if is_reference(node.value):
                    mut = is_mutable(node.scopes, get_id(node.value))
                    fndef.returns.rust_needs_reference = not mut
                    fndef.rust_return_needs_reference = (
                        fndef.returns.rust_needs_reference
                    )
        return node



import ast

from py2many.clike import CLikeTranspiler as CommonCLikeTranspiler


dart_type_map = {
    "bool": "bool",
    "int": "int",
    "float": "double",
    "bytes": "Uint8List",
    "str": "String",
    "c_int8": "int",
    "c_int16": "int",
    "c_int32": "int",
    "c_int64": "int",
    "c_uint8": "int",
    "c_uint16": "int",
    "c_uint32": "int",
    "c_uint64": "int",
}

# allowed as names in Python but treated as keywords in Dart
dart_keywords = frozenset(
    [
        "assert",
        "break",
        "case",
        "catch",
        "class",
        "const",
        "continue",
        "default",
        "do",
        "else",
        "enum",
        "extends",
        "false",
        "final",
        "finally",
        "for",
        "if",
        "in",
        "is",
        "new",
        "null",
        "rethrow",
        "return",
        "super",
        "switch",
        "this",
        "throw",
        "true",
        "try",
        "var",
        "void",
        "while",
        "with",
    ]
)


class CLikeTranspiler(CommonCLikeTranspiler):
    def __init__(self):
        super().__init__()
        self._type_map = dart_type_map

    def visit_Name(self, node):
        if node.id in dart_keywords:
            return node.id + "_"
        return super().visit_Name(node)

    def visit_BinOp(self, node):
        if isinstance(node.op, ast.Pow):
            return "pow({0}, {1})".format(self.visit(node.left), self.visit(node.right))

        left = self.visit(node.left)
        op = self.visit(node.op)
        right = self.visit(node.right)

        # Multiplication and division binds tighter (has higher precedence) than addition and subtraction.
        # To visually communicate this we omit spaces when multiplying and dividing.
        if isinstance(node.op, (ast.Mult, ast.Div)):
            if getattr(node, "use_integer_div", False):
                # Use integer division operator
                op = "~/"
            return f"({left}{op}{right})"
        else:
            return f"({left} {op} {right})"

    def visit_In(self, node):
        left = self.visit(node.left)
        right = self.visit(node.comparators[0])
        return "{0}.any({1})".format(right, left)



import ast
from py2many.scope import add_scope_context
from py2many.context import add_variable_context


def parse(*args):
    source = ast.parse("\n".join(args))
    add_scope_context(source)
    return source


class TestScopeTransformer:
    def test_scope_added(self):
        source = parse("def foo():", "   return 10")
        assert isinstance(source.scopes[-1], ast.Module)
        assert isinstance(source.body[0].scopes[-1], ast.FunctionDef)
        assert isinstance(source.body[0].body[0].scopes[-1], ast.FunctionDef)


class TestScopeList:
    def test_find_returns_most_upper_definition(self):
        source = parse("x = 1", "def foo():", "   x = 2")
        add_variable_context(source)
        definition = source.scopes.find("x")
        assert definition.lineno == 1



import ast

from py2many.analysis import get_id


def detect_mutable_vars(node):
    return MutabilityTransformer().visit(node)


class MutabilityTransformer(ast.NodeTransformer):
    """
    Analyzes every function for mutable variables and put them into FunctionDef node
    """

    def __init__(self):
        self.var_usage_count = {}
        self.lvalue = False

    def increase_use_count(self, name):
        if not name in self.var_usage_count:
            self.var_usage_count[name] = 0
        self.var_usage_count[name] += 1

    def visit_FunctionDef(self, node):
        self.var_usage_count = {}
        self.generic_visit(node)
        mutable_vars = []
        for var, count in self.var_usage_count.items():
            if count > 1:
                mutable_vars.append(var)
        node.mutable_vars = mutable_vars
        return node

    def visit_Assign(self, node):
        old = self.lvalue
        self.lvalue = True
        target = node.targets[0]
        if isinstance(target, ast.Tuple):
            for e in target.elts:
                self.visit(e)
        self.visit(target)
        self.lvalue = old
        self.generic_visit(node)
        return node

    def visit_AugAssign(self, node):
        old = self.lvalue
        self.lvalue = True
        self.visit(node.target)
        self.lvalue = old
        self.generic_visit(node)
        return node

    def visit_AnnAssign(self, node):
        old = self.lvalue
        self.lvalue = True
        self.visit(node.target)
        self.lvalue = old
        self.generic_visit(node)
        return node

    def visit_Subscript(self, node):
        self.visit(node.value)
        self.visit(node.slice)
        return node

    def visit_Name(self, node):
        if self.lvalue:
            self.increase_use_count(get_id(node))
        return node

    def visit_Call(self, node):
        fname = get_id(node.func)
        fndef = node.scopes.find(fname)
        if fndef and hasattr(fndef, "args"):
            for fnarg, node_arg in zip(fndef.args.args, node.args):
                if hasattr(fndef, "mutable_vars") and fnarg.arg in fndef.mutable_vars:
                    self.increase_use_count(get_id(node_arg))
        if hasattr(node.func, "attr"):
            if node.func.attr == "append":
                self.increase_use_count(get_id(node.func.value))
        self.generic_visit(node)
        return node



import sys
import textwrap

from py14.transpiler import transpile


def parse(*args):
    return "\n".join(args)


def test_declare():
    source = parse("x = 3")
    cpp = transpile(source)
    assert cpp == "auto x = 3;"


def test_empty_return():
    source = parse("def foo():", "   return")
    cpp = transpile(source)
    expected = """\
    inline void foo() {
    return;}
    """
    assert cpp == textwrap.dedent(expected)


def test_print_multiple_vars():
    if sys.version_info[0] >= 3:
        source = parse('print(("hi", "there" ))')
    else:
        source = parse('print("hi", "there" )')
    cpp = transpile(source)
    assert cpp == parse(
        'std::cout << std::string{"hi"} << std::string{"there"};',
        "std::cout << std::endl;",
    )


def test_assert():
    source = parse("assert 1 == foo(3)")
    cpp = transpile(source, testing=True)
    assert cpp == '#include "catch.hpp"\nREQUIRE(1 == foo(3));'


def test_augmented_assigns_with_counter():
    source = parse(
        "counter = 0", "counter += 5", "counter -= 2", "counter *= 2", "counter /= 3"
    )
    cpp = transpile(source)
    assert cpp == parse(
        "auto counter = 0;",
        "counter += 5;",
        "counter -= 2;",
        "counter *= 2;",
        "counter /= 3;",
    )


def test_declare_var_before_if_else_statements():
    source = parse("if 10:", "   x = True", "else:", "   x = False", "y = x")
    cpp = transpile(source)
    assert cpp == parse(
        "decltype(true) x;",
        "if(10) {",
        "x = true;",
        "} else {",
        "x = false;",
        "}",
        "auto y = x;",
    )


def test_declare_vars_inside_if_as_long_as_possible():
    source = parse("x = 5", "if 10:", "   y = 10", "   x *= y")
    cpp = transpile(source)
    assert cpp == parse("auto x = 5;", "if(10) {", "auto y = 10;", "x *= y;", "}")


def test_print_program_args():
    source = parse(
        'if __name__ == "__main__":', "    for arg in sys.argv:", "       print(arg)"
    )
    cpp = transpile(source)
    assert cpp == parse(
        "int main(int argc, char ** argv) {",
        "py14::sys::argv = std::vector<std::string>(argv, argv + argc);",
        "for(auto arg : py14::sys::argv) {",
        "std::cout << arg;",
        "std::cout << std::endl;",
        "}}",
        "",
    )


def test_tuple_swap():
    source = parse("x = 3", "y = 1", "x, y = y, x")
    cpp = transpile(source)
    assert cpp == parse(
        "auto x = 3;", "auto y = 1;", "std::tie(x, y) = std::make_tuple(y, x);"
    )


def test_assign():
    source = parse("x = 3", "x = 1")
    cpp = transpile(source)
    assert cpp == parse("auto x = 3;", "x = 1;")


def test_function_with_return():
    source = parse("def fun(x):", "   return x")
    cpp = transpile(source)
    expected = """\
        template <typename T0>auto fun(T0 x) {
        return x;}
    """
    assert cpp == textwrap.dedent(expected)


def test_void_function():
    source = parse("def test_fun():", "   assert True")
    cpp = transpile(source, testing=False)
    expected = """\
        inline void test_fun() {
        assert(true);}
    """
    assert cpp == textwrap.dedent(expected)


def test_create_catch_test_case():
    source = parse("def test_fun():", "   assert True")
    cpp = transpile(source, testing=True)
    assert cpp == parse(
        '#include "catch.hpp"', 'TEST_CASE("test_fun") {', "REQUIRE(true);", "}"
    )


def test_list_as_vector():
    source = parse("values = [0, 1, 2, 3]")
    cpp = transpile(source)
    assert cpp == "std::vector<decltype(0)> values = {0, 1, 2, 3};"


def test_vector_find_out_type():
    source = parse("values = []", "values.append(1)")
    cpp = transpile(source)
    assert cpp == parse("std::vector<decltype(1)> values = {};", "values.push_back(1);")


def test_map_function():
    source = parse(
        "def map(values, fun):",
        "   results = []",
        "   for v in values:",
        "       results.append(fun(v))",
        "   return results",
    )
    cpp = transpile(source)
    expected = """\
        template <typename T0, typename T1>auto map(T0 values, T1 fun) {
        std::vector<decltype(fun(std::declval<typename decltype(values)::value_type>()))> results = {};
        for(auto v : values) {
        results.push_back(fun(v));
        }
        return results;}
    """
    assert cpp == textwrap.dedent(expected)


def test_normal_pdf():
    source = parse(
        "def pdf(x, mean, std_dev):",
        "    term1 = 1.0 / ((2 * math.pi) ** 0.5)",
        "    term2 = (math.e ** (-1.0 * (x-mean) ** 2.0 / 2.0",
        "             * (std_dev ** 2.0)))",
        "    return term1 * term2",
    )
    cpp = transpile(source)
    expected = """\
        template <typename T0, typename T1, typename T2>auto pdf(T0 x, T1 mean, T2 std_dev) {
        auto term1 = 1.0 / (std::pow(2 * (py14::math::pi), 0.5));
        auto term2 = std::pow(py14::math::e, (((-1.0) * (std::pow(x - mean, 2.0))) / 2.0) * (std::pow(std_dev, 2.0)));
        return term1 * term2;}
    """
    assert cpp == parse(textwrap.dedent(expected))
